from pathlib import Path
from typing import Optional, Union, cast

from anchorpy3_core.idl import (
    Idl,
    IdlAccountItem,
    IdlAccounts,
    IdlSeedConst,
    IdlTypeArray,
    IdlTypeSimple,
)
from autoflake import fix_code
from black import FileMode, format_str
from genpy import (
    Assign,
    Collection,
    FromImport,
    If,
    Import,
    ImportAs,
    Line,
    Return,
    Suite,
)
from pyheck import shouty_snake, snake, upper_camel

from anchorpy.clientgen.common import (
    _field_to_encodable,
    _layout_for_type,
    _py_type_from_idl,
    _sanitize,
)
from anchorpy.coder.idl_compat import (
    get_account_writable,
    get_account_signer,
)
from anchorpy.clientgen.genpy_extension import (
    ANNOTATIONS_IMPORT,
    Call,
    Function,
    List,
    NamedArg,
    StrDict,
    StrDictEntry,
    TypedDict,
    TypedParam,
)
from anchorpy.coder.common import _sighash
from anchorpy.coder.idl import FIELD_TYPE_MAP

CONST_ACCOUNTS = {
    "associated_token_program": "ASSOCIATED_TOKEN_PROGRAM_ID",
    "rent": "RENT",
    "system_program": "SYS_PROGRAM_ID",
    "token_program": "TOKEN_PROGRAM_ID",
    "clock": "CLOCK",
}


def gen_instructions(idl: Idl, root: Path, gen_pdas: bool) -> None:
    instructions_dir = root / "instructions"
    instructions_dir.mkdir(exist_ok=True)
    gen_index_file(idl, instructions_dir)
    instructions = gen_instructions_code(idl, instructions_dir, gen_pdas)
    for path, code in instructions.items():
        formatted = format_str(code, mode=FileMode())
        fixed = fix_code(formatted, remove_all_unused_imports=True)
        path.write_text(fixed)


def gen_index_file(idl: Idl, instructions_dir: Path) -> None:
    code = gen_index_code(idl)
    path = instructions_dir / "__init__.py"
    formatted = format_str(code, mode=FileMode())
    path.write_text(formatted)


def gen_index_code(idl: Idl) -> str:
    imports: list[FromImport] = []
    for ix in idl.instructions:
        ix_name_snake_unsanitized = snake(ix.name)
        ix_name = _sanitize(ix_name_snake_unsanitized)
        import_members: list[str] = [ix_name]
        if ix.args:
            import_members.append(_args_interface_name(ix_name_snake_unsanitized))
        if ix.accounts:
            import_members.append(_accounts_interface_name(ix_name_snake_unsanitized))
        if import_members:
            imports.append(FromImport(f".{ix_name}", import_members))
    return str(Collection(imports))


def _args_interface_name(ix_name: str) -> str:
    return f"{upper_camel(ix_name)}Args"


def _accounts_interface_name(ix_name: str) -> str:
    return f"{upper_camel(ix_name)}Accounts"


def recurse_accounts(
    accs: list[IdlAccountItem],
    nested_names: list[str],
    const_accs: dict[int, str],
    acc_idx: int = 0,
) -> tuple[list[str], int]:
    elements: list[str] = []
    for acc in accs:
        names = [*nested_names, _sanitize(snake(acc.name))]
        if isinstance(acc, IdlAccounts):
            nested_accs = cast(IdlAccounts, acc)
            new_elements, acc_idx = recurse_accounts(
                nested_accs.accounts, names, const_accs, acc_idx
            )
            elements.extend(new_elements)
        else:
            acc_idx += 1
            try:
                pubkey_var = const_accs[acc_idx]
            except KeyError:
                try:
                    pubkey_var = CONST_ACCOUNTS[names[-1]]
                except KeyError:
                    nested_keys = [f'["{key}"]' for key in names]
                    dict_accessor = "".join(nested_keys)
                    pubkey_var = f"accounts{dict_accessor}"
            # Get account flags with backward compatibility
            is_signer = get_account_signer(acc)
            is_writable = get_account_writable(acc)

            if acc.is_optional:
                elements.append(
                    f"AccountMeta(pubkey={pubkey_var}, "
                    f"is_signer={is_signer}, "
                    f"is_writable={is_writable}) "
                    f"if {pubkey_var} else "
                    f"AccountMeta(pubkey=program_id, "
                    f"is_signer=False, is_writable=False)"
                )
            else:
                elements.append(
                    f"AccountMeta(pubkey={pubkey_var}, "
                    f"is_signer={is_signer}, "
                    f"is_writable={is_writable})"
                )
    return elements, acc_idx


def to_buffer_value(
    ty: Union[IdlTypeSimple, IdlTypeArray], value: Union[str, int, list[int]]
) -> bytes:
    if isinstance(value, int):
        encoder = FIELD_TYPE_MAP[cast(IdlTypeSimple, ty)]
        return encoder.build(value)
    if isinstance(value, str):
        return value.encode()
    if isinstance(value, list):
        return bytes(value)
    raise ValueError(f"Unexpected type. ty: {ty}; value: {value}")


GenAccountsRes = tuple[list[TypedDict], list[Assign], dict[int, str], int]


def gen_accounts(
    name,
    idl_accs: list[IdlAccountItem],
    gen_pdas: bool,
    accum: Optional[GenAccountsRes] = None,
) -> GenAccountsRes:
    if accum is None:
        extra_typeddicts_to_use: list[TypedDict] = []
        accum_const_pdas: list[Assign] = []
        const_acc_indices: dict[int, str] = {}
        acc_count = 0
    else:
        extra_typeddicts_to_use, accum_const_pdas, const_acc_indices, acc_count = accum
    params: list[TypedParam] = []
    const_pdas: list[Assign] = []
    for acc in idl_accs:
        acc_name = _sanitize(snake(acc.name))
        if isinstance(acc, IdlAccounts):
            nested_accs = cast(IdlAccounts, acc)
            nested_acc_name = f"{upper_camel(nested_accs.name)}Nested"
            nested_res = gen_accounts(
                nested_acc_name,
                nested_accs.accounts,
                gen_pdas,
                (
                    extra_typeddicts_to_use,
                    accum_const_pdas,
                    const_acc_indices,
                    acc_count,
                ),
            )
            if nested_res[0]:
                params.append(TypedParam(acc_name, f"{nested_acc_name}"))
            extra_typeddicts_to_use = extra_typeddicts_to_use + nested_res[0]
            accum_const_pdas = accum_const_pdas + nested_res[1]
            const_acc_indices = const_acc_indices | nested_res[2]
            acc_count = nested_res[3] + 1
        else:
            acc_count += 1
            pda_generated = False
            if gen_pdas:
                maybe_pda = acc.pda
                if maybe_pda is not None and all(
                    isinstance(seed, IdlSeedConst) for seed in maybe_pda.seeds
                ):
                    seeds = cast(list[IdlSeedConst], maybe_pda.seeds)
                    const_pda_name = shouty_snake(f"{name}_{acc_name}")
                    const_pda_body_items = [
                        str(
                            to_buffer_value(
                                cast(Union[IdlTypeSimple, IdlTypeArray], seed.ty),
                                cast(Union[str, int, list[int]], seed.value),
                            )
                        )
                        for seed in seeds
                    ]
                    seeds_arg = List(const_pda_body_items)
                    seeds_named_arg = NamedArg("seeds", seeds_arg)
                    const_pda_body = Call(
                        "Pubkey.find_program_address",
                        [seeds_named_arg, NamedArg("program_id", "PROGRAM_ID")],
                    )
                    const_pdas.append(Assign(const_pda_name, f"{const_pda_body}[0]"))
                    const_acc_indices = {
                        **const_acc_indices,
                        acc_count: const_pda_name,
                    }
                    pda_generated = True
            if not pda_generated:
                try:
                    CONST_ACCOUNTS[acc_name]
                except KeyError:
                    if acc.is_optional:
                        params.append(TypedParam(acc_name, "typing.Optional[Pubkey]"))
                    else:
                        params.append(TypedParam(acc_name, "Pubkey"))
    maybe_typed_dict_container = [TypedDict(name, params)] if params else []
    accounts = maybe_typed_dict_container + extra_typeddicts_to_use
    return accounts, accum_const_pdas + const_pdas, const_acc_indices, acc_count


def gen_instructions_code(idl: Idl, out: Path, gen_pdas: bool) -> dict[Path, str]:
    types_import = [FromImport("..", ["types"])] if idl.types else []
    imports = [
        ANNOTATIONS_IMPORT,
        Import("typing"),
        FromImport("solders.pubkey", ["Pubkey"]),
        FromImport("solders.system_program", ["ID as SYS_PROGRAM_ID"]),
        FromImport("solders.sysvar", ["RENT", "CLOCK"]),
        FromImport(
            "spl.token.constants", ["TOKEN_PROGRAM_ID", "ASSOCIATED_TOKEN_PROGRAM_ID"]
        ),
        FromImport("solders.instruction", ["Instruction", "AccountMeta"]),
        FromImport(
            "anchorpy.borsh_extension", ["BorshPubkey", "EnumForCodegen", "COption"]
        ),
        FromImport("construct", ["Pass", "Construct"]),
        ImportAs("borsh_construct", "borsh"),
        *types_import,
        FromImport("..program_id", ["PROGRAM_ID"]),
    ]
    result = {}
    for ix in idl.instructions:
        ix_name_snake_unsanitized = snake(ix.name)
        ix_name = _sanitize(ix_name_snake_unsanitized)
        filename = (out / ix_name).with_suffix(".py")
        args_interface_params: list[TypedParam] = []
        layout_items: list[str] = []
        encoded_args_entries: list[StrDictEntry] = []
        accounts_interface_name = _accounts_interface_name(ix_name_snake_unsanitized)
        for arg in ix.args:
            arg_name = _sanitize(snake(arg.name))
            args_interface_params.append(
                TypedParam(
                    arg_name,
                    _py_type_from_idl(
                        idl=idl,
                        ty=arg.ty,
                        types_relative_imports=False,
                        use_fields_interface_for_struct=False,
                    ),
                )
            )
            layout_items.append(
                _layout_for_type(
                    idl=idl, ty=arg.ty, name=arg_name, types_relative_imports=False
                )
            )
            encoded_args_entries.append(
                StrDictEntry(
                    arg_name,
                    _field_to_encodable(
                        idl=idl,
                        ty=arg,
                        types_relative_imports=False,
                        val_prefix='args["',
                        val_suffix='"]',
                    ),
                )
            )
        if ix.args:
            args_interface_name = _args_interface_name(ix_name)
            args_interface_container = [
                TypedDict(args_interface_name, args_interface_params)
            ]
            layout_val = f"borsh.CStruct({','.join(layout_items)})"
            layout_assignment_container = [Assign("layout", layout_val)]
            args_container = [TypedParam("args", args_interface_name)]
            encoded_args_val = f"layout.build({StrDict(encoded_args_entries)})"
        else:
            args_interface_container = []
            layout_val = "Pass"
            args_container = []
            layout_assignment_container = []
            encoded_args_val = 'b""'
        accounts_container = (
            [TypedParam("accounts", accounts_interface_name)] if ix.accounts else []
        )
        accounts, const_pdas, const_acc_indices, _ = gen_accounts(
            accounts_interface_name, ix.accounts, gen_pdas
        )
        recursed = recurse_accounts(ix.accounts, [], const_acc_indices)[0]
        keys_assignment = Assign("keys: list[AccountMeta]", f"{List(recursed)}")
        remaining_accounts_concatenation = If(
            "remaining_accounts is not None", Line("keys += remaining_accounts")
        )
        identifier_assignment = Assign(
            "identifier", _sighash(ix_name_snake_unsanitized)
        )
        encoded_args_assignment = Assign("encoded_args", encoded_args_val)
        data_assignment = Assign("data", "identifier + encoded_args")
        returning = Return("Instruction(program_id, data, keys)")
        ix_fn = Function(
            ix_name,
            [
                *args_container,
                *accounts_container,
                TypedParam("program_id", "Pubkey = PROGRAM_ID"),
                TypedParam(
                    "remaining_accounts",
                    "typing.Optional[typing.List[AccountMeta]] = None",
                ),
            ],
            Suite(
                [
                    keys_assignment,
                    remaining_accounts_concatenation,
                    identifier_assignment,
                    encoded_args_assignment,
                    data_assignment,
                    returning,
                ]
            ),
            "Instruction",
        )
        contents = Collection(
            [
                *imports,
                *args_interface_container,
                *layout_assignment_container,
                *const_pdas,
                *accounts,
                ix_fn,
            ]
        )
        result[filename] = str(contents)
    return result



from dataclasses import dataclass
from pathlib import Path
from typing import Union as TypingUnion
from typing import cast

from anchorpy3_core.idl import (
    Idl,
    IdlEnumVariant,
    IdlField,
    IdlType,
    IdlTypeDefinitionTyAlias,
    IdlTypeDefinitionTyStruct,
)
from autoflake import fix_code
from black import FileMode, format_str
from genpy import (
    Assign,
    Collection,
    FromImport,
    Generable,
    If,
    Import,
    ImportAs,
    Raise,
    Return,
    Suite,
)
from pyheck import snake

from anchorpy.clientgen.common import (
    _field_from_decoded,
    _field_from_json,
    _field_to_encodable,
    _field_to_json,
    _idl_type_to_json_type,
    _json_interface_name,
    _kind_interface_name,
    _layout_for_type,
    _py_type_from_idl,
    _sanitize,
    _value_interface_name,
)
from anchorpy.clientgen.genpy_extension import (
    ANNOTATIONS_IMPORT,
    Call,
    ClassMethod,
    Dataclass,
    Function,
    Method,
    NamedArg,
    StrDict,
    StrDictEntry,
    Tuple,
    TupleTypeAlias,
    TypedDict,
    TypedParam,
    Union,
)


def gen_types(idl: Idl, root: Path) -> None:
    types = idl.types
    if types is None or not types:
        return
    types_dir = root / "types"
    types_dir.mkdir(exist_ok=True)
    gen_index_file(idl, types_dir)
    gen_type_files(idl, types_dir)


def gen_index_file(idl: Idl, types_dir: Path) -> None:
    code = gen_index_code(idl)
    path = types_dir / "__init__.py"
    formatted = format_str(code, mode=FileMode())
    path.write_text(formatted)


def gen_index_code(idl: Idl) -> str:
    imports: list[TypingUnion[Import, FromImport]] = [Import("typing")]
    for ty in idl.types:
        ty_type = ty.ty
        sanitized_name = _sanitize(ty.name)
        module_name = _sanitize(snake(ty.name))
        imports.append(FromImport(".", [module_name]))
        if isinstance(ty_type, IdlTypeDefinitionTyStruct):
            import_members = [sanitized_name, _json_interface_name(sanitized_name)]
        elif isinstance(ty_type, IdlTypeDefinitionTyAlias):
            import_members = [sanitized_name]
        else:
            import_members = [
                _kind_interface_name(sanitized_name),
                _json_interface_name(sanitized_name),
            ]
        imports.append(
            FromImport(
                f".{module_name}",
                import_members,
            )
        )
    return str(Collection(imports))


def gen_type_files(idl: Idl, types_dir: Path) -> None:
    types_code = gen_types_code(idl, types_dir)
    for path, code in types_code.items():
        formatted = format_str(code, mode=FileMode())
        fixed = fix_code(formatted, remove_all_unused_imports=True)
        path.write_text(fixed)


def gen_types_code(idl: Idl, out: Path) -> dict[Path, str]:
    res = {}
    types_module_names = [_sanitize(snake(ty.name)) for ty in idl.types]
    for ty in idl.types:
        ty_name = _sanitize(ty.name)
        module_name = _sanitize(snake(ty.name))
        relative_import_items = [
            mod for mod in types_module_names if mod != module_name
        ]
        relative_import_container = (
            [FromImport(".", relative_import_items)] if relative_import_items else []
        )
        ty_type = ty.ty
        if isinstance(ty_type, IdlTypeDefinitionTyAlias):
            body = Assign(
                ty_name,
                _py_type_from_idl(
                    idl,
                    ty_type.value,
                    types_relative_imports=True,
                    use_fields_interface_for_struct=False,
                ),
            )
        elif isinstance(ty_type, IdlTypeDefinitionTyStruct):
            body = gen_struct(idl, ty_name, ty_type.fields)
        else:
            body = gen_enum(idl, ty_name, ty_type.variants)
        code = str(Collection([ANNOTATIONS_IMPORT, *relative_import_container, body]))
        path = (out / module_name).with_suffix(".py")
        res[path] = code
    return res


def gen_struct(idl: Idl, name: str, fields: list[IdlField]) -> Collection:
    imports = [
        Import("typing"),
        FromImport("dataclasses", ["dataclass"]),
        FromImport("construct", ["Container", "Construct"]),
        FromImport("solders.pubkey", ["Pubkey"]),
        FromImport("anchorpy.borsh_extension", ["BorshPubkey"]),
        ImportAs("borsh_construct", "borsh"),
    ]
    json_interface_name = _json_interface_name(name)
    field_params: list[TypedParam] = []
    json_interface_params: list[TypedParam] = []
    layout_items: list[str] = []
    from_decoded_items: list[str] = []
    to_encodable_items: list[str] = []
    to_json_items: list[str] = []
    from_json_items: list[str] = []
    for field in fields:
        field_name = _sanitize(snake(field.name))
        field_params.append(
            TypedParam(
                field_name,
                _py_type_from_idl(
                    idl=idl,
                    ty=field.ty,
                    types_relative_imports=True,
                    use_fields_interface_for_struct=False,
                ),
            )
        )
        json_interface_params.append(
            TypedParam(
                field_name,
                _idl_type_to_json_type(ty=field.ty, types_relative_imports=True),
            )
        )
        layout_items.append(
            _layout_for_type(
                idl=idl, ty=field.ty, name=field_name, types_relative_imports=True
            )
        )
        from_decoded_item_val = _field_from_decoded(
            idl=idl,
            ty=IdlField(name=snake(field.name), docs=None, ty=field.ty),
            val_prefix="obj.",
            types_relative_imports=True,
        )
        from_decoded_items.append(f"{field_name}={from_decoded_item_val}")
        as_encodable = _field_to_encodable(
            idl=idl,
            ty=field,
            val_prefix="self.",
            val_suffix="",
            types_relative_imports=True,
        )
        to_encodable_items.append(f'"{field_name}": {as_encodable}')
        to_json_items.append(f'"{field_name}": {_field_to_json(idl, field, "self.")}')
        field_from_json = _field_from_json(
            idl=idl, ty=field, types_relative_imports=True
        )
        from_json_items.append(f"{field_name}={field_from_json}")
    json_interface = TypedDict(json_interface_name, json_interface_params)
    layout = f"borsh.CStruct({','.join(layout_items)})"
    args_for_from_decoded = ",".join(from_decoded_items)
    to_encodable_body = "{" + ",".join(to_encodable_items) + "}"
    to_json_body = "{" + ",".join(to_json_items) + "}"
    args_for_from_json = ",".join(from_json_items)
    struct_cls = Dataclass(
        name,
        [
            Assign("layout: typing.ClassVar", layout),
            *field_params,
            ClassMethod(
                "from_decoded",
                [TypedParam("obj", "Container")],
                Return(f"cls({args_for_from_decoded})"),
                f'"{name}"',
            ),
            Method(
                "to_encodable",
                [],
                Return(to_encodable_body),
                "dict[str, typing.Any]",
            ),
            Method("to_json", [], Return(to_json_body), json_interface_name),
            ClassMethod(
                "from_json",
                [TypedParam("obj", json_interface_name)],
                Return(f"cls({args_for_from_json})"),
                f'"{name}"',
            ),
        ],
    )
    return Collection([*imports, json_interface, struct_cls])


def _make_cstruct(fields: dict[str, str]) -> str:
    formatted_fields = ",".join([f'"{key}" / {val}' for key, val in fields.items()])
    return f"borsh.CStruct({formatted_fields})"


@dataclass
class _NamedFieldRecord:
    field_type_alias_entry: TypedParam
    value_type_alias_entry: TypedParam
    json_interface_value_type_entry: TypedParam
    json_value_item: StrDictEntry
    encodable_value_item: StrDictEntry
    init_entry_for_from_decoded: NamedArg
    init_entry_for_from_json: NamedArg


def _make_named_field_record(
    named_field: IdlField, idl: Idl, cast_obj_var_name: str
) -> _NamedFieldRecord:
    named_field_name = _sanitize(snake(named_field.name))
    return _NamedFieldRecord(
        field_type_alias_entry=TypedParam(
            named_field_name,
            _py_type_from_idl(
                idl=idl,
                ty=named_field.ty,
                types_relative_imports=True,
                use_fields_interface_for_struct=False,
            ),
        ),
        value_type_alias_entry=TypedParam(
            named_field_name,
            _py_type_from_idl(
                idl=idl,
                ty=named_field.ty,
                types_relative_imports=True,
                use_fields_interface_for_struct=False,
            ),
        ),
        json_interface_value_type_entry=TypedParam(
            named_field_name,
            _idl_type_to_json_type(ty=named_field.ty, types_relative_imports=True),
        ),
        json_value_item=StrDictEntry(
            named_field_name,
            _field_to_json(idl, named_field, 'self.value["', val_suffix='"]'),
        ),
        encodable_value_item=StrDictEntry(
            named_field_name,
            _field_to_encodable(
                idl=idl,
                ty=named_field,
                val_prefix='self.value["',
                val_suffix='"]',
                types_relative_imports=True,
            ),
        ),
        init_entry_for_from_decoded=NamedArg(
            named_field_name,
            _field_from_decoded(
                idl=idl,
                ty=IdlField(f'val["{named_field_name}"]', docs=None, ty=named_field.ty),
                types_relative_imports=True,
                val_prefix="",
            ),
        ),
        init_entry_for_from_json=NamedArg(
            named_field_name,
            _field_from_json(
                idl=idl,
                ty=named_field,
                param_prefix=f'{cast_obj_var_name}["',
                param_suffix='"]',
                types_relative_imports=True,
            ),
        ),
    )


@dataclass
class _UnnamedFieldRecord:
    field_type_alias_element: str
    value_type_alias_element: str
    json_interface_value_element: str
    json_value_element: str
    encodable_value_item: StrDictEntry
    init_element_for_from_decoded: str
    init_element_for_from_json: str


def _make_unnamed_field_record(
    index: int, unnamed_field: IdlType, idl: Idl, cast_obj_var_name: str
) -> _UnnamedFieldRecord:
    elem_name = f"value[{index}]"
    encodable = _field_to_encodable(
        idl=idl,
        ty=IdlField(f"[{index}]", docs=None, ty=unnamed_field),
        val_prefix="self.value",
        types_relative_imports=True,
        convert_case=False,
    )
    return _UnnamedFieldRecord(
        field_type_alias_element=_py_type_from_idl(
            idl=idl,
            ty=unnamed_field,
            types_relative_imports=True,
            use_fields_interface_for_struct=False,
        ),
        value_type_alias_element=_py_type_from_idl(
            idl=idl,
            ty=unnamed_field,
            types_relative_imports=True,
            use_fields_interface_for_struct=False,
        ),
        json_interface_value_element=_idl_type_to_json_type(
            ty=unnamed_field, types_relative_imports=True
        ),
        json_value_element=_field_to_json(
            idl,
            IdlField(elem_name, docs=None, ty=unnamed_field),
            "self.",
            convert_case=False,
        ),
        encodable_value_item=StrDictEntry(f"item_{index}", encodable),
        init_element_for_from_decoded=_field_from_decoded(
            idl=idl,
            ty=IdlField(f'val["item_{index}"]', docs=None, ty=unnamed_field),
            val_prefix="",
            types_relative_imports=True,
        ),
        init_element_for_from_json=_field_from_json(
            idl=idl,
            ty=IdlField(str(index), docs=None, ty=unnamed_field),
            param_prefix=f"{cast_obj_var_name}[",
            param_suffix="]",
            types_relative_imports=True,
        ),
    )


def gen_enum(idl: Idl, name: str, variants: list[IdlEnumVariant]) -> Collection:
    imports = [
        Import("typing"),
        FromImport("dataclasses", ["dataclass"]),
        FromImport("solders.pubkey", ["Pubkey"]),
        FromImport("construct", ["Construct"]),
        FromImport("anchorpy.borsh_extension", ["EnumForCodegen", "BorshPubkey"]),
        ImportAs("borsh_construct", "borsh"),
    ]
    invalid_enum_raise = Raise('ValueError("Invalid enum object")')
    from_decoded_dict_check = If("not isinstance(obj, dict)", invalid_enum_raise)
    variant_name_in_obj_checks: list[Generable] = []
    obj_kind_checks: list[Generable] = []
    json_interfaces: list[TypedDict] = []
    classes: list[Dataclass] = []
    cstructs: list[str] = []
    type_variants_members: list[str] = []
    json_variants_members: list[str] = []
    json_interface_value_field_types: list[TypingUnion[TypedDict, TupleTypeAlias]] = []
    value_type_aliases: list[TypingUnion[TypedDict, TupleTypeAlias]] = []
    for idx, variant in enumerate(variants):
        discriminator = idx
        fields = variant.fields
        variant_name = _sanitize(variant.name)
        value_interface_name = _value_interface_name(variant.name)
        json_interface_name = _json_interface_name(variant.name)
        json_interface_value_type_name = f"{json_interface_name}Value"
        cast_obj_var_name = snake(json_interface_value_type_name)
        encodable_value_items: list[StrDictEntry] = []
        cstruct_fields: dict[str, str] = {}
        to_json_params_base = NamedArg("kind", f'"{variant.name}"')
        json_interface_kind_field = TypedParam(
            "kind", f'typing.Literal["{variant.name}"]'
        )
        type_variants_members.append(variant_name)
        json_variants_members.append(_json_interface_name(variant.name))

        def make_variant_name_in_obj_check(then: Generable) -> Generable:
            return If(f'"{variant.name}" in obj', then)

        def make_obj_kind_check(return_val: str) -> Generable:
            accessed = 'obj["kind"]'
            lhs = f"{accessed} == "
            cast_obj_assignment_container = (
                [
                    Assign(
                        cast_obj_var_name,
                        f'typing.cast({json_interface_value_type_name}, obj["value"])',
                    )
                ]
                if fields
                else []
            )
            return If(
                f'{lhs}"{variant.name}"',
                Suite([*cast_obj_assignment_container, Return(return_val)]),
            )

        if fields is not None:
            flds = fields.fields
            val_line_for_from_decoded = Assign("val", f'obj["{variant.name}"]')
            if isinstance(flds[0], IdlField):
                named_enum_fields = cast(list[IdlField], flds)
                value_type_alias_entries: list[TypedParam] = []
                json_interface_value_type_entries: list[TypedParam] = []
                json_value_items: list[StrDictEntry] = []
                init_entries_for_from_decoded: list[NamedArg] = []
                init_entries_for_from_json: list[NamedArg] = []
                for named_field in named_enum_fields:
                    rec = _make_named_field_record(named_field, idl, cast_obj_var_name)
                    value_type_alias_entries.append(rec.value_type_alias_entry)
                    json_interface_value_type_entries.append(
                        rec.json_interface_value_type_entry
                    )
                    json_value_items.append(rec.json_value_item)
                    encodable_value_items.append(rec.encodable_value_item)
                    init_entries_for_from_decoded.append(
                        rec.init_entry_for_from_decoded
                    )
                    init_entries_for_from_json.append(rec.init_entry_for_from_json)

                    cstruct_fields[snake(named_field.name)] = _layout_for_type(
                        idl=idl, ty=named_field.ty, types_relative_imports=True
                    )
                value_type_aliases.append(
                    TypedDict(value_interface_name, value_type_alias_entries)
                )
                json_interface_value_field_type = TypedDict(
                    json_interface_value_type_name, json_interface_value_type_entries
                )
                json_value_entry = StrDict(json_value_items)
                to_json_body = Return(
                    Call(
                        json_interface_name,
                        [to_json_params_base, NamedArg("value", str(json_value_entry))],
                    )
                )
                init_arg_for_from_decoded = Call(
                    value_interface_name, init_entries_for_from_decoded
                )
                init_arg_for_from_json = Call(
                    value_interface_name, init_entries_for_from_json
                )
            else:
                tuple_enum_fields = cast(list[IdlType], flds)
                field_type_alias_elements: list[str] = []
                value_type_alias_elements: list[str] = []
                json_interface_value_elements: list[str] = []
                json_value_elements: list[str] = []
                init_elements_for_from_decoded: list[str] = []
                init_elements_for_from_json: list[str] = []
                for i, unnamed_field in enumerate(tuple_enum_fields):
                    rec_unnamed = _make_unnamed_field_record(
                        i, unnamed_field, idl, cast_obj_var_name
                    )
                    field_type_alias_elements.append(
                        rec_unnamed.field_type_alias_element
                    )
                    value_type_alias_elements.append(
                        rec_unnamed.value_type_alias_element
                    )
                    json_interface_value_elements.append(
                        rec_unnamed.json_interface_value_element
                    )
                    json_value_elements.append(rec_unnamed.json_value_element)
                    encodable_value_items.append(rec_unnamed.encodable_value_item)
                    init_elements_for_from_decoded.append(
                        rec_unnamed.init_element_for_from_decoded
                    )
                    init_elements_for_from_json.append(
                        rec_unnamed.init_element_for_from_json
                    )
                    cstruct_fields[f"item_{i}"] = _layout_for_type(
                        idl=idl, ty=unnamed_field, types_relative_imports=True
                    )
                value_type_aliases.append(
                    TupleTypeAlias(
                        value_interface_name,
                        value_type_alias_elements,
                    )
                )
                json_interface_value_field_type = TupleTypeAlias(
                    json_interface_value_type_name, json_interface_value_elements
                )
                to_json_body = Return(
                    Call(
                        json_interface_name,
                        [
                            to_json_params_base,
                            NamedArg("value", str(Tuple(json_value_elements))),
                        ],
                    )
                )
                init_arg_for_from_decoded = Tuple(init_elements_for_from_decoded)
                init_arg_for_from_json = Tuple(init_elements_for_from_json)
            value_field_container = [TypedParam("value", value_interface_name)]
            to_json_method = Method("to_json", [], to_json_body, json_interface_name)
            encodable_value_entry = StrDict(encodable_value_items)
            to_encodable_body = Return(
                StrDict([StrDictEntry(variant.name, encodable_value_entry)])
            )
            to_encodable_method = Method("to_encodable", [], to_encodable_body, "dict")
            json_interface_params = [
                TypedParam("value", json_interface_value_type_name),
                json_interface_kind_field,
            ]
            json_interface_value_field_types.append(json_interface_value_field_type)
            variant_name_in_obj_check = make_variant_name_in_obj_check(
                Suite(
                    [
                        val_line_for_from_decoded,
                        Return(
                            f"{_sanitize(variant.name)}"
                            f"({init_arg_for_from_decoded})"
                        ),
                    ]
                )
            )
            obj_kind_check = make_obj_kind_check(
                f"{_sanitize(variant.name)}({init_arg_for_from_json})",
            )
        else:
            to_json_method = ClassMethod(
                "to_json",
                [],
                Return(Call(json_interface_name, [to_json_params_base])),
                json_interface_name,
            )
            to_encodable_method = ClassMethod(
                "to_encodable",
                [],
                Return(StrDict([StrDictEntry(variant.name, StrDict([]))])),
                "dict",
            )
            value_field_container = []
            json_interface_params = [json_interface_kind_field]
            variant_name_in_obj_check = make_variant_name_in_obj_check(
                Return(f"{_sanitize(variant.name)}()")
            )
            obj_kind_check = make_obj_kind_check(f"{_sanitize(variant.name)}()")
        json_interfaces.append(TypedDict(json_interface_name, json_interface_params))
        class_common_attrs = [
            Assign("discriminator: typing.ClassVar", discriminator),
            Assign("kind: typing.ClassVar", f'"{variant.name}"'),
        ]

        attrs = [
            *class_common_attrs,
            *value_field_container,
            to_json_method,
            to_encodable_method,
        ]
        classes.append(Dataclass(_sanitize(variant.name), attrs))
        variant_name_in_obj_checks.append(variant_name_in_obj_check)
        obj_kind_checks.append(obj_kind_check)
        cstructs.append(f'"{variant.name}" / {_make_cstruct(cstruct_fields)}')
    from_decoded_fn = Function(
        "from_decoded",
        [TypedParam("obj", "dict")],
        Suite(
            [from_decoded_dict_check, *variant_name_in_obj_checks, invalid_enum_raise]
        ),
        _kind_interface_name(name),
    )
    from_json_fn = Function(
        "from_json",
        [TypedParam("obj", _json_interface_name(name))],
        Suite(
            [
                *obj_kind_checks,
                Assign("kind", 'obj["kind"]'),
                Raise('ValueError(f"Unrecognized enum kind: {kind}")'),
            ]
        ),
        _kind_interface_name(name),
    )
    formatted_cstructs = ",".join(cstructs)
    layout_assignment = Assign(
        "layout",
        f"EnumForCodegen({formatted_cstructs})",
    )
    json_variants = Union(json_variants_members)
    type_variants = Union(type_variants_members)
    kind_type_alias = Assign(_kind_interface_name(name), type_variants)
    json_type_alias = Assign(_json_interface_name(name), json_variants)
    return Collection(
        [
            *imports,
            *json_interface_value_field_types,
            *value_type_aliases,
            *json_interfaces,
            *classes,
            kind_type_alias,
            json_type_alias,
            from_decoded_fn,
            from_json_fn,
            layout_assignment,
        ]
    )



#!/usr/bin/env python3
"""Test full loopscale_v2.json piece by piece."""

import json
from anchorpy3_core.idl import Idl

with open('../loopscale_v2.json') as f:
    v2_data = json.load(f)

# Build up the IDL piece by piece
test_idl = {
    "version": v2_data.get("version"),
    "name": v2_data.get("name"),
    "address": v2_data.get("address"),
    "metadata": v2_data.get("metadata"),
    "instructions": [],
    "accounts": [],
    "types": [],
    "errors": [],
    "events": []
}

print("Testing loopscale_v2.json components...")
print("="*50)

# Test 1: Base structure
print("\n1. Base structure only")
try:
    idl = Idl.from_json(json.dumps(test_idl))
    print("   ✅ Base structure works")
except Exception as e:
    print(f"   ❌ Base structure failed: {e}")
    exit(1)

# Test 2: Add errors
print("\n2. Adding errors")
test_idl["errors"] = v2_data.get("errors", [])
try:
    idl = Idl.from_json(json.dumps(test_idl))
    print(f"   ✅ Errors work ({len(test_idl['errors'])} errors)")
except Exception as e:
    print(f"   ❌ Errors failed: {e}")
    exit(1)

# Test 3: Add accounts
print("\n3. Adding accounts")
test_idl["accounts"] = v2_data.get("accounts", [])
try:
    idl = Idl.from_json(json.dumps(test_idl))
    print(f"   ✅ Accounts work ({len(test_idl['accounts'])} accounts)")
except Exception as e:
    print(f"   ❌ Accounts failed: {e}")
    exit(1)

# Test 4: Add types
print("\n4. Adding types")
test_idl["types"] = v2_data.get("types", [])
try:
    idl = Idl.from_json(json.dumps(test_idl))
    print(f"   ✅ Types work ({len(test_idl['types'])} types)")
except Exception as e:
    print(f"   ❌ Types failed: {e}")
    # Try to find which type is problematic
    print("\n   Finding problematic type...")
    for i in range(len(v2_data["types"])):
        test_idl["types"] = v2_data["types"][:i+1]
        try:
            Idl.from_json(json.dumps(test_idl))
        except Exception as e2:
            print(f"   Type {i} ({v2_data['types'][i].get('name')}) causes issue")
            print(f"   Error: {e2}")
            print(f"   Type structure: {json.dumps(v2_data['types'][i], indent=2)}")
            break
    exit(1)

# Test 5: Add events
print("\n5. Adding events")
test_idl["events"] = v2_data.get("events", [])
try:
    idl = Idl.from_json(json.dumps(test_idl))
    print(f"   ✅ Events work ({len(test_idl['events'])} events)")
except Exception as e:
    print(f"   ❌ Events failed: {e}")
    exit(1)

# Test 6: Add instructions
print("\n6. Adding instructions")
test_idl["instructions"] = v2_data.get("instructions", [])
try:
    idl = Idl.from_json(json.dumps(test_idl))
    print(f"   ✅ Instructions work ({len(test_idl['instructions'])} instructions)")
except Exception as e:
    print(f"   ❌ Instructions failed: {e}")
    exit(1)

print("\n" + "="*50)
print("✅ All components work individually!")
print("\nTesting full IDL...")

# Test the full IDL
try:
    idl = Idl.from_json(json.dumps(v2_data))
    print("✅ Full IDL parsed successfully!")
except Exception as e:
    print(f"❌ Full IDL failed: {e}")


"""This module provides the `localnet_fixture` fixture factory."""
import os
import signal
import subprocess
from pathlib import Path
from typing import AsyncGenerator, Callable, Literal, Optional, Union

from pytest import fixture
from pytest_asyncio import fixture as async_fixture
from xprocess import ProcessStarter, XProcess, XProcessInfo

from anchorpy.program.core import Program
from anchorpy.workspace import close_workspace, create_workspace

_Scope = Literal["session", "package", "module", "class", "function"]


class _FixedXProcessInfo(XProcessInfo):
    def terminate(self, timeout=60):  # noqa: ARG002
        if not self.pid:
            return 0
        try:
            pgid = os.getpgid(self.pid)
        except ProcessLookupError:
            return 0
        try:
            os.killpg(pgid, signal.SIGTERM)
        except OSError as err:
            print(f"Error while terminating process {err}")
            return -1
        return 1


class _FixedXProcess(XProcess):
    def getinfo(self, name: str) -> _FixedXProcessInfo:
        """Return Process Info for the given external process.

        Args:
            name: Name of the external process.
        """
        return _FixedXProcessInfo(self.rootdir, name)

    def ensure(
        self, name: str, preparefunc: ProcessStarter, restart: bool = False
    ) -> tuple:
        """Return (PID, logfile) from a newly started or already running process.

        Args:
            name: Name of the external process, used for caching info across test runs.
            preparefunc: A subclass of ProcessStarter.
            restart: Force restarting the process if it is running.

        Raises:
            RuntimeError: If process fails to start within the required time.

        Returns:
            (PID, logfile) logfile will be seeked to the end if the
            server was running, otherwise seeked to the line after
            where the wait pattern matched.
        """
        info = self.getinfo(name)
        if not restart and not info.isrunning():
            restart = True

        if restart:
            # ensure the process is terminated first
            if info.pid is not None:
                info.terminate()

            controldir = info.controldir.ensure(dir=1)
            starter = preparefunc(controldir, self)
            args = [str(x) for x in starter.args]
            self.log.debug("%s$ %s", controldir, " ".join(args))
            stdout = open(str(info.logpath), "wb", 0)  # noqa: SIM115

            # is env still necessary? we could pass all in popen_kwargs
            kwargs = {"env": starter.env}

            popen_kwargs = {
                "stdout": stdout,
                "stderr": subprocess.STDOUT,
                # this gives the user the ability to
                # override the previous keywords if
                # desired
                **starter.popen_kwargs,
            }

            kwargs["close_fds"] = True

            # keep references of all popen
            # and info objects for cleanup
            self._info_objects.append((info, starter.terminate_on_interrupt))
            popen_instance = subprocess.Popen(
                args,
                **popen_kwargs,
                **kwargs,  # type: ignore
            )
            self._popen_instances.append(popen_instance)

            info.pid = pid = self._popen_instances[-1].pid
            info.pidpath.write(str(pid))
            self.log.debug("process %r started pid=%s", name, pid)
            stdout.close()

        # keep track of all file handles so we can
        # cleanup later during teardown phase
        self._file_handles.append(info.logpath.open())

        if not restart:
            self._file_handles[-1].seek(0, 2)
        else:
            if not starter.wait(self._file_handles[-1]):
                raise RuntimeError(
                    f"Could not start process {name}, the specified "
                    f"log pattern was not found within {starter.max_read_lines} lines."
                )
            self.log.debug("%s process startup detected", name)

        pytest_extlogfiles = self.config.__dict__.setdefault("_extlogfiles", {})
        pytest_extlogfiles[name] = self._file_handles[-1]
        self.getinfo(name)

        return info.pid, info.logpath


@async_fixture(scope="session")
def _fixed_xprocess(request):
    """Yield session-scoped XProcess helper to manage long-running processes required for testing."""  # noqa: E501
    rootdir = request.config.rootdir
    with _FixedXProcess(request.config, rootdir) as xproc:
        # pass in xprocess object into pytest_unconfigure
        # through config for proper cleanup during teardown
        request.config._xprocess = xproc
        yield xproc


def localnet_fixture(
    path: Path,
    scope: _Scope = "module",
    timeout_seconds: int = 60,
    build_cmd: Optional[str] = None,
) -> Callable:
    """Create a fixture that sets up and tears down a localnet instance with workspace programs deployed.

    Args:
        path: Path to root of the Anchor project.
        scope: Pytest fixture scope.
        timeout_seconds: Time to wait for Anchor localnet to start.
        build_cmd: Command to run before `anchor localnet`. Defaults to `anchor build`.

    Returns:
        A localnet fixture for use with pytest.
    """  # noqa: E501,D202

    @fixture(scope=scope)
    def _localnet_fixture(_fixed_xprocess):
        class Starter(ProcessStarter):
            # startup pattern
            pattern = "JSON RPC URL"
            terminate_on_interrupt = True
            # command to start process
            args = ["anchor", "localnet", "--skip-build"]
            timeout = timeout_seconds
            popen_kwargs = {
                "cwd": path,
                "start_new_session": True,
            }
            max_read_lines = 1_000
            # command to start process

        actual_build_cmd = "anchor build" if build_cmd is None else build_cmd
        subprocess.run(actual_build_cmd, cwd=path, check=True, shell=True)
        # ensure process is running and return its logfile
        logfile = _fixed_xprocess.ensure("localnet", Starter)

        yield logfile

        # clean up whole process tree afterwards
        _fixed_xprocess.getinfo("localnet").terminate()

    return _localnet_fixture


# Should figure out how to reuse localnet_fixture
# instead of copy-pasting (Pytest didn't like it).
def workspace_fixture(
    path: Union[Path, str],
    scope: _Scope = "module",
    timeout_seconds: int = 60,
    build_cmd: Optional[str] = None,
) -> Callable:
    """Create a fixture that sets up and tears down a localnet instance and returns a workspace dict.

    Equivalent to combining `localnet_fixture`, `create_workspace` and `close_workspace`.

    Args:
        path: Path to root of the Anchor project.
        scope: Pytest fixture scope.
        timeout_seconds: Time to wait for Anchor localnet to start.
        build_cmd: Command to run before `anchor localnet`. Defaults to `anchor build`.

    Returns:
        A workspace fixture for use with pytest.
    """  # noqa: E501,D202

    @async_fixture(scope=scope)
    async def _workspace_fixture(
        _fixed_xprocess,
    ) -> AsyncGenerator[dict[str, Program], None]:
        class Starter(ProcessStarter):
            # startup pattern
            pattern = "JSON RPC URL"
            terminate_on_interrupt = True
            # command to start process
            args = ["anchor", "localnet", "--skip-build"]
            timeout = timeout_seconds
            popen_kwargs = {
                "cwd": path,
                "start_new_session": True,
            }
            max_read_lines = 1_000
            # command to start process

        actual_build_cmd = "anchor build" if build_cmd is None else build_cmd
        subprocess.run(actual_build_cmd, cwd=path, check=True, shell=True)
        # ensure process is running
        _ = _fixed_xprocess.ensure("localnet", Starter)
        ws = create_workspace(path)
        yield ws
        await close_workspace(ws)

        # clean up whole process tree afterwards
        _fixed_xprocess.getinfo("localnet").terminate()

    return _workspace_fixture



"""This module deals with generating program instructions."""
from typing import Any, Callable, Sequence, Tuple, cast

from anchorpy3_core.idl import IdlAccount, IdlAccountItem, IdlAccounts, IdlInstruction
from pyheck import snake
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from anchorpy.program.common import (
    NamedInstruction,
    _to_instruction,
    validate_accounts,
)
from anchorpy.program.context import (
    EMPTY_CONTEXT,
    Accounts,
    Context,
    _check_args_length,
)


class _InstructionFn:
    """Callable object to create a `Instruction` generated from an IDL.

    Additionally it provides an `accounts` utility method, returning a list
    of ordered accounts for the instruction.
    """

    def __init__(
        self,
        idl_ix: IdlInstruction,
        encode_fn: Callable[[NamedInstruction], bytes],
        program_id: Pubkey,
    ) -> None:
        """Init.

        Args:
            idl_ix: IDL instruction object
            encode_fn: [description]
            program_id: The program ID.

        Raises:
            ValueError: [description]
        """
        if snake(idl_ix.name) == "_inner":
            raise ValueError("_inner name is reserved")
        self.idl_ix = idl_ix
        self.encode_fn = encode_fn
        self.program_id = program_id

    def __call__(
        self,
        *args: Any,
        ctx: Context = EMPTY_CONTEXT,
    ) -> Instruction:
        """Create the Instruction.

        Args:
            *args: The positional arguments for the program. The type and number
                of these arguments depend on the program being used.
            ctx: non-argument parameters to pass to the method.
        """
        _check_args_length(self.idl_ix, args)
        validate_accounts(self.idl_ix.accounts, ctx.accounts)
        _validate_instruction(self.idl_ix, args)

        keys = self.accounts(ctx.accounts)
        if ctx.remaining_accounts:
            keys.extend(ctx.remaining_accounts)
        return Instruction(
            accounts=keys,
            program_id=self.program_id,
            data=self.encode_fn(_to_instruction(self.idl_ix, args)),
        )

    def accounts(self, accs: Accounts) -> list[AccountMeta]:
        """Order the accounts for this instruction.

        Args:
            accs: Accounts from `ctx` kwarg.

        Returns:
            Ordered and flattened accounts.
        """
        return _accounts_array(accs, self.idl_ix.accounts)


def _accounts_array(
    ctx: Accounts,
    accounts: Sequence[IdlAccountItem],
) -> list[AccountMeta]:
    """Create a list of AccountMeta from a (possibly nested) dict of accounts.

    Args:
        ctx: `accounts` field from the `Context` object.
        accounts: accounts from the IDL.

    Returns:
        AccountMeta objects.
    """
    accounts_ret: list[AccountMeta] = []
    for acc in accounts:
        if isinstance(acc, IdlAccounts):
            rpc_accs = cast(Accounts, ctx[snake(acc.name)])
            acc_arr = _accounts_array(rpc_accs, acc.accounts)
            accounts_ret.extend(acc_arr)
        else:
            account: IdlAccount = acc
            single_account = cast(Pubkey, ctx[snake(account.name)])
            accounts_ret.append(
                AccountMeta(
                    pubkey=single_account,
                    is_writable=account.is_mut,
                    is_signer=account.is_signer,
                ),
            )
    return accounts_ret


def _validate_instruction(ix: IdlInstruction, args: Tuple):  # noqa: ARG001
    """Throws error if any argument required for the `ix` is not given.

    Args:
        ix: The IDL instruction object.
        args: The instruction arguments.
    """
    # TODO: this isn't implemented in the TS client yet



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from solders.instruction import Instruction, AccountMeta
from construct import Construct
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class InitializeWithValues2Args(typing.TypedDict):
    vec_of_option: list[typing.Optional[int]]


layout = borsh.CStruct(
    "vec_of_option" / borsh.Vec(typing.cast(Construct, borsh.Option(borsh.U64)))
)


class InitializeWithValues2Accounts(typing.TypedDict):
    state: Pubkey
    payer: Pubkey


def initialize_with_values2(
    args: InitializeWithValues2Args,
    accounts: InitializeWithValues2Accounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=accounts["payer"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xf8\xbe\x15a\xef\x94'\xb5"
    encoded_args = layout.build(
        {
            "vec_of_option": args["vec_of_option"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from .initialize import initialize, InitializeAccounts
from .initialize_with_values import (
    initialize_with_values,
    InitializeWithValuesArgs,
    InitializeWithValuesAccounts,
)
from .initialize_with_values2 import (
    initialize_with_values2,
    InitializeWithValues2Args,
    InitializeWithValues2Accounts,
)
from .cause_error import cause_error
from .init_my_account import init_my_account, InitMyAccountArgs, InitMyAccountAccounts
from .increment_state_when_present import (
    increment_state_when_present,
    IncrementStateWhenPresentAccounts,
)
from .type_alias import type_alias, TypeAliasArgs



import asyncio
import json
from pathlib import Path
from typing import AsyncGenerator, List

from anchorpy import Provider, Wallet
from anchorpy.pytest_plugin import localnet_fixture
from construct import ListContainer
from pytest import fixture, mark
from pytest_asyncio import fixture as async_fixture
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Confirmed, Processed
from solana.rpc.core import RPCException
from solders.hash import Hash
from solders.keypair import Keypair
from solders.message import Message
from solders.pubkey import Pubkey
from solders.transaction import VersionedTransaction

from tests.client_gen.example_program_gen.accounts import State, State2
from tests.client_gen.example_program_gen.errors import from_tx_error
from tests.client_gen.example_program_gen.errors.custom import SomeError
from tests.client_gen.example_program_gen.instructions import (
    InitializeWithValuesAccounts,
    InitializeWithValuesArgs,
    cause_error,
    increment_state_when_present,
    initialize,
    initialize_with_values,
    initialize_with_values2,
)
from tests.client_gen.example_program_gen.program_id import PROGRAM_ID
from tests.client_gen.example_program_gen.types import BarStruct, FooStruct
from tests.client_gen.example_program_gen.types.foo_enum import (
    Named,
    NoFields,
    Struct,
    Unnamed,
)

EXAMPLE_PROGRAM_DIR = Path("tests/client_gen/example-program")


@fixture(scope="session")
def event_loop():
    """Create an instance of the default event loop for each test case."""
    loop = asyncio.get_event_loop_policy().new_event_loop()
    yield loop
    loop.close()


localnet = localnet_fixture(EXAMPLE_PROGRAM_DIR, scope="session")


@fixture(scope="session")
def payer(localnet) -> Keypair:
    with (EXAMPLE_PROGRAM_DIR / ".anchor/test-ledger/faucet-keypair.json").open() as f:
        faucet_keypair_json: List[int] = json.load(f)
    return Keypair.from_bytes(faucet_keypair_json)


@async_fixture(scope="session")
async def provider(localnet, payer: Keypair) -> AsyncGenerator[Provider, None]:
    wallet = Wallet(payer)
    conn = AsyncClient(commitment=Processed)
    prov = Provider(conn, wallet)
    yield prov
    await prov.close()


@async_fixture(scope="session")
async def blockhash(provider: Provider) -> Hash:
    return (await provider.connection.get_latest_blockhash(Confirmed)).value.blockhash


@async_fixture(scope="module")
async def init_and_account_fetch(provider: Provider, blockhash: Hash) -> Keypair:
    state = Keypair()
    initialize_ix = initialize(
        {
            "state": state.pubkey(),
            "payer": provider.wallet.public_key,
        }
    )
    msg = Message.new_with_blockhash(
        [initialize_ix], provider.wallet.public_key, blockhash
    )
    tx = VersionedTransaction(msg, [provider.wallet.payer, state])
    await provider.send(tx)
    return state


@mark.asyncio
async def test_init_and_account_fetch(
    init_and_account_fetch: Keypair, provider: Provider
) -> None:
    state = init_and_account_fetch
    vec_struct_field_enum_field_expected = Named(
        value={
            "bool_field": True,
            "u8_field": 15,
            "nested": BarStruct(some_field=True, other_field=10),
        }
    )
    assert vec_struct_field_enum_field_expected.discriminator == 2
    assert vec_struct_field_enum_field_expected.kind == "Named"
    vec_struct_field_expected = [
        FooStruct(
            field1=123,
            field2=999,
            nested=BarStruct(some_field=True, other_field=10),
            vec_nested=[BarStruct(some_field=True, other_field=10)],
            option_nested=BarStruct(some_field=True, other_field=10),
            enum_field=vec_struct_field_enum_field_expected,
        )
    ]
    option_struct_field_expected = FooStruct(
        field1=123,
        field2=999,
        nested=BarStruct(some_field=True, other_field=10),
        vec_nested=[BarStruct(some_field=True, other_field=10)],
        option_nested=BarStruct(some_field=True, other_field=10),
        enum_field=Named(
            value={
                "bool_field": True,
                "u8_field": 15,
                "nested": BarStruct(some_field=True, other_field=10),
            }
        ),
    )
    struct_field_expected = FooStruct(
        field1=123,
        field2=999,
        nested=BarStruct(some_field=True, other_field=10),
        vec_nested=[BarStruct(some_field=True, other_field=10)],
        option_nested=BarStruct(some_field=True, other_field=10),
        enum_field=Named(
            value={
                "bool_field": True,
                "u8_field": 15,
                "nested": BarStruct(some_field=True, other_field=10),
            }
        ),
    )
    array_field_expected = ListContainer([True, False, True])
    vec_field_expected = ListContainer([1, 2, 100, 1000, 18446744073709551615])
    enum_field1_expected = Unnamed(
        value=(False, 10, BarStruct(some_field=True, other_field=10))
    )
    assert enum_field1_expected.kind == "Unnamed"
    assert enum_field1_expected.discriminator == 0
    enum_field2_expected = Named(
        value={
            "bool_field": True,
            "u8_field": 20,
            "nested": BarStruct(some_field=True, other_field=10),
        }
    )
    assert enum_field2_expected.kind == "Named"
    assert enum_field2_expected.discriminator == 2
    enum_field3_expected = Struct(value=(BarStruct(some_field=True, other_field=10),))
    assert enum_field3_expected.discriminator == 3
    assert enum_field3_expected.kind == "Struct"
    enum_field4_expected = NoFields()
    assert enum_field4_expected.discriminator == 6
    assert enum_field4_expected.kind == "NoFields"
    expected = State(
        bool_field=True,
        u8_field=234,
        i8_field=-123,
        u16_field=62345,
        i16_field=-31234,
        u32_field=1234567891,
        i32_field=-1234567891,
        f32_field=123456.5,
        u64_field=9223372036854775817,
        i64_field=-4611686018427387914,
        f64_field=1234567891.345,
        u128_field=170141183460469231731687303715884105737,
        i128_field=-85070591730234615865843651857942052874,
        bytes_field=b"\x01\x02\xff\xfe",
        string_field="hello",
        pubkey_field=Pubkey.from_string("EPZP2wrcRtMxrAPJCXVEQaYD9eH7fH7h12YqKDcd4aS7"),
        vec_field=vec_field_expected,
        vec_struct_field=vec_struct_field_expected,
        option_field=None,
        option_struct_field=option_struct_field_expected,
        struct_field=struct_field_expected,
        array_field=array_field_expected,
        enum_field1=enum_field1_expected,
        enum_field2=enum_field2_expected,
        enum_field3=enum_field3_expected,
        enum_field4=enum_field4_expected,
    )
    res = await State.fetch(provider.connection, state.pubkey())
    assert res == expected
    res = await State.fetch(provider.connection, state.pubkey(), program_id=PROGRAM_ID)
    assert res == expected


@async_fixture(scope="session")
async def setup_fetch_multiple(
    provider: Provider, blockhash: Hash
) -> tuple[Keypair, Keypair]:
    state = Keypair()
    another_state = Keypair()
    initialize_ixs = [
        initialize(
            {
                "state": state.pubkey(),
                "payer": provider.wallet.public_key,
            }
        ),
        initialize(
            {
                "state": another_state.pubkey(),
                "payer": provider.wallet.public_key,
            }
        ),
    ]
    msg = Message.new_with_blockhash(
        initialize_ixs, provider.wallet.public_key, blockhash
    )
    tx = VersionedTransaction(msg, [provider.wallet.payer, state, another_state])
    await provider.send(tx)
    return state, another_state


@mark.asyncio
async def test_fetch_multiple(
    provider: Provider, setup_fetch_multiple: tuple[Keypair, Keypair]
) -> None:
    state, another_state = setup_fetch_multiple
    non_state = Keypair()
    res = await State.fetch_multiple(
        provider.connection,
        [state.pubkey(), non_state.pubkey(), another_state.pubkey()],
    )
    assert isinstance(res[0], State)
    assert res[1] is None
    assert isinstance(res[2], State)


@async_fixture(scope="session")
async def send_instructions_with_args(
    provider: Provider, blockhash
) -> tuple[Keypair, Keypair]:
    state = Keypair()
    state2 = Keypair()
    vec_struct_field = [
        FooStruct(
            field1=1,
            field2=2,
            nested=BarStruct(some_field=True, other_field=55),
            vec_nested=[BarStruct(some_field=False, other_field=11)],
            option_nested=None,
            enum_field=Unnamed((True, 22, BarStruct(some_field=True, other_field=33))),
        )
    ]
    struct_field = FooStruct(
        field1=1,
        field2=2,
        nested=BarStruct(some_field=True, other_field=55),
        vec_nested=[BarStruct(some_field=False, other_field=11)],
        option_nested=None,
        enum_field=NoFields(),
    )
    enum_field1 = Unnamed((True, 15, BarStruct(some_field=False, other_field=200)))
    enum_field2 = Named(
        {
            "bool_field": True,
            "u8_field": 128,
            "nested": BarStruct(some_field=False, other_field=1),
        }
    )
    enum_field3 = Struct((BarStruct(some_field=True, other_field=15),))
    initialize_with_values_args = InitializeWithValuesArgs(
        bool_field=True,
        u8_field=253,
        i8_field=-120,
        u16_field=61234,
        i16_field=-31253,
        u32_field=1234567899,
        i32_field=-123456789,
        f32_field=123458.5,
        u64_field=9223372036854775810,
        i64_field=-4611686018427387912,
        f64_field=1234567892.445,
        u128_field=170141183460469231731687303715884105740,
        i128_field=-85070591730234615865843651857942052877,
        bytes_field=bytes([5, 10, 255]),
        string_field="string value",
        pubkey_field=Pubkey.from_string("GDddEKTjLBqhskzSMYph5o54VYLQfPCR3PoFqKHLJK6s"),
        vec_field=[1, 123456789123456789],
        vec_struct_field=vec_struct_field,
        option_field=True,
        option_struct_field=None,
        struct_field=struct_field,
        array_field=[True, True, False],
        enum_field1=enum_field1,
        enum_field2=enum_field2,
        enum_field3=enum_field3,
        enum_field4=NoFields(),
    )
    initialize_with_values_accounts = InitializeWithValuesAccounts(
        state=state.pubkey(),
        payer=provider.wallet.public_key,
    )
    ix1 = initialize_with_values(
        initialize_with_values_args, initialize_with_values_accounts
    )
    ix2 = initialize_with_values2(
        {"vec_of_option": [None, None, 20]},
        {
            "state": state2.pubkey(),
            "payer": provider.wallet.public_key,
        },
    )
    msg = Message.new_with_blockhash([ix1, ix2], provider.wallet.public_key, blockhash)
    tx = VersionedTransaction(msg, [provider.wallet.payer, state, state2])
    await provider.send(tx)
    return state, state2


@mark.asyncio
async def test_instructions_with_args(
    send_instructions_with_args: tuple[Keypair, Keypair], provider: Provider
) -> None:
    state, state2 = send_instructions_with_args
    expected = State(
        bool_field=True,
        u8_field=253,
        i8_field=-120,
        u16_field=61234,
        i16_field=-31253,
        u32_field=1234567899,
        i32_field=-123456789,
        f32_field=123458.5,
        u64_field=9223372036854775810,
        i64_field=-4611686018427387912,
        f64_field=1234567892.445,
        u128_field=170141183460469231731687303715884105740,
        i128_field=-85070591730234615865843651857942052877,
        bytes_field=b"\x05\n\xff",
        string_field="string value",
        pubkey_field=Pubkey.from_string("GDddEKTjLBqhskzSMYph5o54VYLQfPCR3PoFqKHLJK6s"),
        vec_field=ListContainer([1, 123456789123456789]),
        vec_struct_field=[
            FooStruct(
                field1=1,
                field2=2,
                nested=BarStruct(some_field=True, other_field=55),
                vec_nested=[BarStruct(some_field=False, other_field=11)],
                option_nested=None,
                enum_field=Unnamed(
                    value=(True, 22, BarStruct(some_field=True, other_field=33))
                ),
            )
        ],
        option_field=True,
        option_struct_field=None,
        struct_field=FooStruct(
            field1=1,
            field2=2,
            nested=BarStruct(some_field=True, other_field=55),
            vec_nested=[BarStruct(some_field=False, other_field=11)],
            option_nested=None,
            enum_field=NoFields(),
        ),
        array_field=ListContainer([True, True, False]),
        enum_field1=Unnamed(
            value=(True, 15, BarStruct(some_field=False, other_field=200))
        ),
        enum_field2=Named(
            value={
                "bool_field": True,
                "u8_field": 128,
                "nested": BarStruct(some_field=False, other_field=1),
            }
        ),
        enum_field3=Struct(value=(BarStruct(some_field=True, other_field=15),)),
        enum_field4=NoFields(),
    )
    expected2 = State2(vec_of_option=ListContainer([None, None, 20]))
    res = await State.fetch(provider.connection, state.pubkey())
    res2 = await State2.fetch(provider.connection, state2.pubkey())
    assert res == expected
    assert res2 == expected2


@mark.asyncio
async def test_instruction_with_optional_account(
    send_instructions_with_args: tuple[Keypair, Keypair],
    provider: Provider,
    blockhash: Hash,
) -> None:
    def prepare_tx(ixs):
        msg = Message.new_with_blockhash([ix], provider.wallet.public_key, blockhash)
        tx = VersionedTransaction(msg, [provider.wallet.payer])
        return tx

    state, state2 = send_instructions_with_args
    ix = increment_state_when_present(
        {
            "first_state": None,
            "second_state": state2.pubkey(),
        },
    )
    before_res = await State.fetch(provider.connection, state.pubkey())
    assert before_res is not None
    tx = prepare_tx(ix)
    await provider.send(tx)

    res = await State.fetch(provider.connection, state.pubkey())
    assert res is not None
    assert before_res.u8_field == res.u8_field

    ix = increment_state_when_present(
        {
            "first_state": state.pubkey(),
            "second_state": state2.pubkey(),
        },
    )

    res = await State.fetch(provider.connection, state.pubkey())
    tx = prepare_tx(ix)
    await provider.send(tx)

    res = await State.fetch(provider.connection, state.pubkey())
    assert res is not None
    assert before_res.u8_field + 1 == res.u8_field


@mark.asyncio
async def test_cause_error(provider: Provider, blockhash: Hash) -> None:
    msg = Message.new_with_blockhash(
        [cause_error()], provider.wallet.public_key, blockhash
    )
    tx = VersionedTransaction(msg, [provider.wallet.payer])
    try:
        await provider.send(tx)
    except RPCException as exc:
        caught = from_tx_error(exc)
        assert isinstance(caught, SomeError)



from pathlib import Path
from typing import cast

from anchorpy3_core.idl import (
    Idl,
    IdlField,
    IdlTypeDefinition,
    IdlTypeDefinitionTyStruct,
)
from autoflake import fix_code
from black import FileMode, format_str
from genpy import (
    Assign,
    Collection,
    For,
    FromImport,
    If,
    Import,
    ImportAs,
    Raise,
    Return,
    Statement,
    Suite,
)
from pyheck import snake

from anchorpy.clientgen.common import (
    _field_from_decoded,
    _field_from_json,
    _field_to_json,
    _idl_type_to_json_type,
    _json_interface_name,
    _layout_for_type,
    _py_type_from_idl,
    _sanitize,
)
from anchorpy.clientgen.genpy_extension import (
    Call,
    ClassMethod,
    Continue,
    Dataclass,
    Method,
    NamedArg,
    StrDict,
    StrDictEntry,
    TypedDict,
    TypedParam,
)
from anchorpy.coder.accounts import _account_discriminator
from anchorpy.coder.idl_compat import get_account_type_definition


def gen_accounts(idl: Idl, root: Path) -> None:
    accounts = idl.accounts
    if accounts is None or not accounts:
        return
    accounts_dir = root / "accounts"
    accounts_dir.mkdir(exist_ok=True)
    gen_index_file(idl, accounts_dir)
    accounts_dict = gen_accounts_code(idl, accounts_dir)
    for path, code in accounts_dict.items():
        formatted = format_str(code, mode=FileMode())
        fixed = fix_code(formatted, remove_all_unused_imports=True)
        path.write_text(fixed)


def gen_index_file(idl: Idl, accounts_dir: Path) -> None:
    code = gen_index_code(idl)
    formatted = format_str(code, mode=FileMode())
    (accounts_dir / "__init__.py").write_text(formatted)


def gen_index_code(idl: Idl) -> str:
    imports: list[FromImport] = []
    for acc in idl.accounts:
        # Handle both object and dict formats
        name = acc.name if hasattr(acc, 'name') else acc.get('name')
        acc_name = _sanitize(name)
        members = [
            acc_name,
            _json_interface_name(acc_name),
        ]
        module_name = _sanitize(snake(name))
        imports.append(FromImport(f".{module_name}", members))
    return str(Collection(imports))


def gen_accounts_code(idl: Idl, accounts_dir: Path) -> dict[Path, str]:
    res = {}
    for acc in idl.accounts:
        # Handle both object and dict formats
        name = acc.name if hasattr(acc, 'name') else acc.get('name')
        filename = f"{_sanitize(snake(name))}.py"
        path = accounts_dir / filename
        code = gen_account_code(acc, idl)
        res[path] = code
    return res


def gen_account_code(acc, idl: Idl) -> str:
    base_imports = [
        Import("typing"),
        FromImport("dataclasses", ["dataclass"]),
        FromImport("construct", ["Construct"]),
        FromImport("solders.pubkey", ["Pubkey"]),
        FromImport("solana.rpc.async_api", ["AsyncClient"]),
        FromImport("solana.rpc.commitment", ["Commitment"]),
        ImportAs("borsh_construct", "borsh"),
        FromImport("anchorpy.coder.accounts", ["ACCOUNT_DISCRIMINATOR_SIZE"]),
        FromImport("anchorpy.error", ["AccountInvalidDiscriminator"]),
        FromImport("anchorpy.utils.rpc", ["get_multiple_accounts"]),
        FromImport(
            "anchorpy.borsh_extension", ["BorshPubkey", "EnumForCodegen", "COption"]
        ),
        FromImport("..program_id", ["PROGRAM_ID"]),
    ]
    imports = (
        [*base_imports, FromImport("..", ["types"])] if idl.types else base_imports
    )
    fields_interface_params: list[TypedParam] = []
    json_interface_params: list[TypedParam] = []

    # Get account name (handle both object and dict)
    acc_name_raw = acc.name if hasattr(acc, 'name') else acc.get('name')
    name = _sanitize(acc_name_raw)

    # For old format, the account IS the type definition
    # For new format, we need to look it up in types array
    # Check if account has a 'ty' or 'type' field (old format)
    has_ty_field = hasattr(acc, 'ty') or (isinstance(acc, dict) and ('ty' in acc or 'type' in acc))

    if has_ty_field:
        # Old format: account is the type definition
        type_def = acc
    else:
        # New format: need to look up in types array
        type_def = get_account_type_definition(acc, idl.types)

    # Get the struct type
    if hasattr(type_def, 'ty'):
        ty = cast(IdlTypeDefinitionTyStruct, type_def.ty)
    elif isinstance(type_def, dict) and 'ty' in type_def:
        ty = type_def['ty']
    elif isinstance(type_def, dict) and 'type' in type_def:
        ty = type_def['type']
    else:
        # Assume the type_def itself is the type structure
        ty = type_def

    # Get fields
    if hasattr(ty, 'fields'):
        fields = ty.fields
    elif isinstance(ty, dict) and 'fields' in ty:
        fields = ty['fields']
    else:
        fields = []
    json_interface_name = _json_interface_name(name)
    layout_items: list[str] = []
    init_body_assignments: list[Assign] = []
    decode_body_entries: list[NamedArg] = []
    to_json_entries: list[StrDictEntry] = []
    from_json_entries: list[NamedArg] = []
    for field in fields:
        # Handle both object and dict field formats
        if hasattr(field, 'name'):
            field_name_raw = field.name
            field_ty = field.ty
        elif isinstance(field, dict):
            field_name_raw = field.get('name')
            field_ty = field.get('type') or field.get('ty')
        else:
            # Skip if we can't determine the field structure
            continue

        field_name = _sanitize(snake(field_name_raw))
        fields_interface_params.append(
            TypedParam(
                field_name,
                _py_type_from_idl(
                    idl=idl,
                    ty=field_ty,
                    types_relative_imports=False,
                    use_fields_interface_for_struct=False,
                ),
            )
        )
        json_interface_params.append(
            TypedParam(
                field_name,
                _idl_type_to_json_type(ty=field_ty, types_relative_imports=False),
            )
        )
        layout_items.append(
            _layout_for_type(
                idl=idl, ty=field_ty, name=field_name, types_relative_imports=False
            )
        )
        init_body_assignments.append(
            Assign(f"self.{field_name}", f'fields["{field_name}"]')
        )
        decode_body_entries.append(
            NamedArg(
                field_name,
                _field_from_decoded(
                    idl=idl,
                    ty=IdlField(name=snake(field_name_raw), docs=None, ty=field_ty),
                    types_relative_imports=False,
                    val_prefix="dec.",
                ),
            )
        )
        # Create a proper field object for _field_to_json
        field_obj = IdlField(name=field_name_raw, docs=None, ty=field_ty) if isinstance(field, dict) else field
        to_json_entries.append(
            StrDictEntry(field_name, _field_to_json(idl, field_obj, "self."))
        )
        from_json_entries.append(
            NamedArg(
                field_name,
                _field_from_json(idl=idl, ty=field_obj, types_relative_imports=False),
            )
        )
    json_interface = TypedDict(json_interface_name, json_interface_params)
    discriminator_assignment = Assign(
        "discriminator: typing.ClassVar", _account_discriminator(name)
    )
    layout_assignment = Assign(
        "layout: typing.ClassVar", f"borsh.CStruct({','.join(layout_items)})"
    )
    fetch_method = ClassMethod(
        "fetch",
        [
            TypedParam("conn", "AsyncClient"),
            TypedParam("address", "Pubkey"),
            TypedParam("commitment", "typing.Optional[Commitment] = None"),
            TypedParam("program_id", "Pubkey = PROGRAM_ID"),
        ],
        Suite(
            [
                Assign(
                    "resp",
                    "await conn.get_account_info(address, commitment=commitment)",
                ),
                Assign("info", "resp.value"),
                If("info is None", Return("None")),
                If(
                    "info.owner != program_id",
                    Raise('ValueError("Account does not belong to this program")'),
                ),
                Assign("bytes_data", "info.data"),
                Return("cls.decode(bytes_data)"),
            ]
        ),
        f'typing.Optional["{name}"]',
        is_async=True,
    )
    account_does_not_belong_raise = Raise(
        'ValueError("Account does not belong to this program")'
    )
    fetch_multiple_return_type = f'typing.List[typing.Optional["{name}"]]'
    fetch_multiple_method = ClassMethod(
        "fetch_multiple",
        [
            TypedParam("conn", "AsyncClient"),
            TypedParam("addresses", "list[Pubkey]"),
            TypedParam("commitment", "typing.Optional[Commitment] = None"),
            TypedParam("program_id", "Pubkey = PROGRAM_ID"),
        ],
        Suite(
            [
                Assign(
                    "infos",
                    (
                        "await get_multiple_accounts"
                        "(conn, addresses,commitment=commitment)"
                    ),
                ),
                Assign(f"res: {fetch_multiple_return_type}", "[]"),
                For(
                    "info",
                    "infos",
                    Suite(
                        [
                            If(
                                "info is None",
                                Suite([Statement("res.append(None)"), Continue()]),
                            ),
                            If(
                                "info.account.owner != program_id",
                                account_does_not_belong_raise,
                            ),
                            Statement("res.append(cls.decode(info.account.data))"),
                        ]
                    ),
                ),
                Return("res"),
            ]
        ),
        f'typing.List[typing.Optional["{name}"]]',
        is_async=True,
    )
    decode_body_end = Call("cls", decode_body_entries)
    account_invalid_raise = Raise(
        'AccountInvalidDiscriminator("The discriminator for this account is invalid")'
    )
    decode_method = ClassMethod(
        "decode",
        [TypedParam("data", "bytes")],
        Suite(
            [
                If(
                    "data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator",
                    account_invalid_raise,
                ),
                Assign(
                    "dec", f"{name}.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])"
                ),
                Return(decode_body_end),
            ]
        ),
        f'"{name}"',
    )
    to_json_body = StrDict(to_json_entries)
    to_json_method = Method("to_json", [], Return(to_json_body), json_interface_name)
    from_json_body = Call("cls", from_json_entries)
    from_json_method = ClassMethod(
        "from_json",
        [TypedParam("obj", json_interface_name)],
        Return(from_json_body),
        f'"{name}"',
    )
    klass = Dataclass(
        name,
        [
            discriminator_assignment,
            layout_assignment,
            *fields_interface_params,
            fetch_method,
            fetch_multiple_method,
            decode_method,
            to_json_method,
            from_json_method,
        ],
    )
    return str(
        Collection(
            [
                *imports,
                json_interface,
                klass,
            ]
        )
    )



#!/usr/bin/env python3
"""Comprehensive test of all IDL formats with AnchorPy."""

import json
import tempfile
from pathlib import Path
from anchorpy3_core.idl import Idl
from anchorpy import Program, Provider
from anchorpy.clientgen.accounts import gen_accounts
from anchorpy.clientgen.errors import gen_errors
from anchorpy.clientgen.instructions import gen_instructions
from anchorpy.clientgen.program_id import gen_program_id
from anchorpy.clientgen.types import gen_types
import sys

def test_idl_comprehensive(idl_name: str, description: str):
    """Comprehensive test of an IDL file."""
    print(f"\n{'='*70}")
    print(f"Testing: {idl_name}")
    print(f"Format: {description}")
    print(f"{'='*70}")

    # Load the IDL
    idl_path = Path(__file__).parent.parent / idl_name
    print(f"📁 Loading from: {idl_path}")

    with open(idl_path) as f:
        idl_json = json.load(f)

    # Test 1: Parse as Idl object
    print("\n1️⃣  IDL Parsing...")
    try:
        idl_obj = Idl.from_json(json.dumps(idl_json))
        print(f"   ✅ Successfully parsed IDL")

        # Display IDL stats
        print(f"\n   📊 IDL Statistics:")
        print(f"      Version: {idl_json.get('version', 'None')}")
        print(f"      Address: {idl_json.get('address', 'None')}")

        if idl_obj.metadata:
            spec = getattr(idl_obj.metadata, 'spec', None) if hasattr(idl_obj.metadata, 'spec') else (
                idl_obj.metadata.get('spec') if isinstance(idl_obj.metadata, dict) else None
            )
            if spec:
                print(f"      Spec: {spec}")

        print(f"      Accounts: {len(idl_obj.accounts) if idl_obj.accounts else 0}")
        print(f"      Instructions: {len(idl_obj.instructions) if idl_obj.instructions else 0}")
        print(f"      Types: {len(idl_obj.types) if idl_obj.types else 0}")
        print(f"      Events: {len(idl_obj.events) if idl_obj.events else 0}")
        print(f"      Errors: {len(idl_obj.errors) if idl_obj.errors else 0}")

        # Check format
        if idl_obj.accounts and len(idl_obj.accounts) > 0:
            first_acc = idl_obj.accounts[0]
            if hasattr(first_acc, 'discriminator') or (isinstance(first_acc, dict) and 'discriminator' in first_acc):
                print(f"      Format: New (v0.1.0+) - has discriminators")
            else:
                print(f"      Format: Old (legacy)")

    except Exception as e:
        print(f"   ❌ Failed to parse IDL: {e}")
        return False

    # Test 2: Create Program object
    print("\n2️⃣  Program Creation...")
    try:
        provider = Provider.local()
        program_id = idl_json.get('address', 'So11111111111111111111111111111111111111112')
        program = Program(idl_json, program_id, provider)
        print(f"   ✅ Successfully created Program object")
        print(f"      Program ID: {program.program_id}")

        # Check available methods
        if hasattr(program, 'rpc'):
            rpc_methods = dir(program.rpc)
            user_methods = [m for m in rpc_methods if not m.startswith('_')]
            print(f"      RPC methods available: {len(user_methods)}")
            if user_methods:
                print(f"      Sample methods: {', '.join(user_methods[:3])}")

        if hasattr(program, 'account'):
            account_types = dir(program.account)
            user_accounts = [a for a in account_types if not a.startswith('_')]
            print(f"      Account types available: {len(user_accounts)}")
            if user_accounts:
                print(f"      Sample accounts: {', '.join(user_accounts[:3])}")

    except Exception as e:
        print(f"   ❌ Failed to create Program: {e}")
        return False

    # Test 3: Generate client code
    print("\n3️⃣  Client Code Generation...")
    program_id = idl_json.get('address')
    if not program_id:
        metadata = idl_json.get('metadata', {})
        program_id = metadata.get('address')
    if not program_id:
        program_id = 'So11111111111111111111111111111111111111112'

    with tempfile.TemporaryDirectory() as tmpdir:
        output_path = Path(tmpdir) / "generated"
        output_path.mkdir()

        generation_steps = [
            ("program_id.py", lambda: gen_program_id(program_id, output_path)),
            ("errors", lambda: gen_errors(idl_obj, output_path)),
            ("instructions", lambda: gen_instructions(idl_obj, output_path, gen_pdas=False)),
            ("types", lambda: gen_types(idl_obj, output_path)),
            ("accounts", lambda: gen_accounts(idl_obj, output_path))
        ]

        all_success = True
        for step_name, step_func in generation_steps:
            try:
                step_func()
                print(f"   ✅ Generated {step_name}")
            except Exception as e:
                print(f"   ❌ Failed to generate {step_name}: {e}")
                all_success = False
                break

        if all_success:
            # Count generated files
            account_files = list((output_path / "accounts").glob("*.py")) if (output_path / "accounts").exists() else []
            instruction_files = list((output_path / "instructions").glob("*.py")) if (output_path / "instructions").exists() else []
            type_files = list((output_path / "types").glob("*.py")) if (output_path / "types").exists() else []
            error_files = list((output_path / "errors").glob("*.py")) if (output_path / "errors").exists() else []

            print(f"\n   📦 Generated Files Summary:")
            print(f"      Account files: {len(account_files)}")
            print(f"      Instruction files: {len(instruction_files)}")
            print(f"      Type files: {len(type_files)}")
            print(f"      Error files: {len(error_files)}")
            print(f"      Total: {len(account_files) + len(instruction_files) + len(type_files) + len(error_files) + 1} files")

            # Sample generated content validation
            if account_files:
                sample_account = account_files[0].read_text()
                if "class " in sample_account and "def fetch" in sample_account:
                    print(f"   ✅ Account files contain proper class definitions")

            if instruction_files:
                sample_ix = instruction_files[0].read_text()
                if "def " in sample_ix:
                    print(f"   ✅ Instruction files contain function definitions")

            return True
        else:
            return False

# Test all IDLs
idls = [
    ("kamino_lend_v4.json", "Old format (legacy)"),
    ("adrena.json", "Old format (legacy)"),
    ("loopscale_v1.json", "New format (v0.1.0)"),
    ("loopscale_v2.json", "New format (v0.1.0 with events)")
]

results = []
for idl_file, description in idls:
    success = test_idl_comprehensive(idl_file, description)
    results.append((idl_file, success))

# Print final summary
print(f"\n{'='*70}")
print("FINAL SUMMARY")
print(f"{'='*70}")
print(f"{'IDL File':<30} {'Status':<10} {'Result'}")
print(f"{'-'*30} {'-'*10} {'-'*20}")

for idl_file, success in results:
    status = "✅ PASSED" if success else "❌ FAILED"
    result = "Full support" if success else "Issues found"
    print(f"{idl_file:<30} {status:<10} {result}")

# Overall result
all_passed = all(success for _, success in results)
print(f"\n{'='*70}")
if all_passed:
    print("🎉 SUCCESS: All IDLs are fully supported!")
    print("\nAnchorPy now supports:")
    print("  ✅ Old format IDLs (legacy Anchor)")
    print("  ✅ New format IDLs (v0.1.0 spec)")
    print("  ✅ Full backward compatibility")
    print("  ✅ Complete client code generation")
    print("  ✅ Program object creation")
    sys.exit(0)
else:
    print("⚠️ PARTIAL SUCCESS: Some IDLs have issues")
    sys.exit(1)


from anchorpy import Coder, Idl
from anchorpy.program.context import Context
from anchorpy.program.namespace.instruction import _InstructionFn
from anchorpy.program.namespace.transaction import _build_transaction_fn
from pytest import fixture
from solders.hash import Hash
from solders.instruction import Instruction
from solders.keypair import Keypair
from solders.pubkey import Pubkey

DEFAULT_PUBKEY = Pubkey.default()


def _make_ix(data: bytes) -> Instruction:
    return Instruction(
        accounts=[],
        program_id=DEFAULT_PUBKEY,
        data=data,
    )


@fixture
def pre_ix() -> Instruction:
    return _make_ix(b"pre")


@fixture
def post_ix() -> Instruction:
    return _make_ix(b"post")


@fixture
def idl() -> Idl:
    raw = """{
        "version": "0.0.0",
        "name": "basic_0",
        "instructions": [
            {
                "name": "initialize",
                "accounts": [],
                "args": []
            }
        ]
    }"""
    return Idl.from_json(raw)


@fixture
def coder(idl: Idl) -> Coder:
    return Coder(idl)


def test_pre_instructions(coder: Coder, idl: Idl, pre_ix: Instruction) -> None:
    coder.instruction.encode
    ix_item = _InstructionFn(
        idl.instructions[0], coder.instruction.build, DEFAULT_PUBKEY
    )
    tx_item = _build_transaction_fn(idl.instructions[0], ix_item)
    tx = tx_item(
        ctx=Context(pre_instructions=[pre_ix]),
        payer=Keypair(),
        blockhash=Hash.default(),
    )
    assert len(tx.message.instructions) == 2


def test_post_instructions(coder: Coder, idl: Idl, post_ix: Instruction) -> None:
    coder.instruction.encode
    ix_item = _InstructionFn(
        idl.instructions[0], coder.instruction.build, DEFAULT_PUBKEY
    )
    tx_item = _build_transaction_fn(idl.instructions[0], ix_item)
    tx = tx_item(
        ctx=Context(post_instructions=[post_ix]),
        payer=Keypair(),
        blockhash=Hash.default(),
    )
    assert len(tx.message.instructions) == 2



"""Pytest config."""
import asyncio
import subprocess
from pathlib import Path

from pytest import fixture

# Since our other fixtures have module scope, we need to define
# this event_loop fixture and give it module scope otherwise
# pytest-asyncio will break.


@fixture(scope="module")
def event_loop():
    """Create an instance of the default event loop for each test case."""
    loop = asyncio.get_event_loop_policy().new_event_loop()
    yield loop
    loop.close()


@fixture(scope="session")
def project_parent_dir(tmpdir_factory) -> Path:
    return Path(tmpdir_factory.mktemp("temp"))


@fixture(scope="session")
def project_dir(project_parent_dir: Path) -> Path:
    proj_dir = project_parent_dir / "tmp"
    command = (
        f"anchorpy client-gen tests/idls/clientgen_example_program.json {proj_dir} "
        "--program-id 3rTQ3R4B2PxZrAyx7EUefySPgZY8RhJf16cZajbmrzp8 --pdas"
    )
    subprocess.run(
        command,
        shell=True,
        check=True,
    )
    return proj_dir



from __future__ import annotations
from . import bar_struct, foo_enum
import typing
from dataclasses import dataclass
from construct import Container, Construct
import borsh_construct as borsh


class FooStructJSON(typing.TypedDict):
    field1: int
    field2: int
    nested: bar_struct.BarStructJSON
    vec_nested: list[bar_struct.BarStructJSON]
    option_nested: typing.Optional[bar_struct.BarStructJSON]
    enum_field: foo_enum.FooEnumJSON


@dataclass
class FooStruct:
    layout: typing.ClassVar = borsh.CStruct(
        "field1" / borsh.U8,
        "field2" / borsh.U16,
        "nested" / bar_struct.BarStruct.layout,
        "vec_nested" / borsh.Vec(typing.cast(Construct, bar_struct.BarStruct.layout)),
        "option_nested" / borsh.Option(bar_struct.BarStruct.layout),
        "enum_field" / foo_enum.layout,
    )
    field1: int
    field2: int
    nested: bar_struct.BarStruct
    vec_nested: list[bar_struct.BarStruct]
    option_nested: typing.Optional[bar_struct.BarStruct]
    enum_field: foo_enum.FooEnumKind

    @classmethod
    def from_decoded(cls, obj: Container) -> "FooStruct":
        return cls(
            field1=obj.field1,
            field2=obj.field2,
            nested=bar_struct.BarStruct.from_decoded(obj.nested),
            vec_nested=list(
                map(
                    lambda item: bar_struct.BarStruct.from_decoded(item), obj.vec_nested
                )
            ),
            option_nested=(
                None
                if obj.option_nested is None
                else bar_struct.BarStruct.from_decoded(obj.option_nested)
            ),
            enum_field=foo_enum.from_decoded(obj.enum_field),
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "field1": self.field1,
            "field2": self.field2,
            "nested": self.nested.to_encodable(),
            "vec_nested": list(map(lambda item: item.to_encodable(), self.vec_nested)),
            "option_nested": (
                None
                if self.option_nested is None
                else self.option_nested.to_encodable()
            ),
            "enum_field": self.enum_field.to_encodable(),
        }

    def to_json(self) -> FooStructJSON:
        return {
            "field1": self.field1,
            "field2": self.field2,
            "nested": self.nested.to_json(),
            "vec_nested": list(map(lambda item: item.to_json(), self.vec_nested)),
            "option_nested": (
                None if self.option_nested is None else self.option_nested.to_json()
            ),
            "enum_field": self.enum_field.to_json(),
        }

    @classmethod
    def from_json(cls, obj: FooStructJSON) -> "FooStruct":
        return cls(
            field1=obj["field1"],
            field2=obj["field2"],
            nested=bar_struct.BarStruct.from_json(obj["nested"]),
            vec_nested=list(
                map(
                    lambda item: bar_struct.BarStruct.from_json(item), obj["vec_nested"]
                )
            ),
            option_nested=(
                None
                if obj["option_nested"] is None
                else bar_struct.BarStruct.from_json(obj["option_nested"])
            ),
            enum_field=foo_enum.from_json(obj["enum_field"]),
        )



import typing
from dataclasses import dataclass
from solders.pubkey import Pubkey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey, COption
from ..program_id import PROGRAM_ID
from .. import types


class AccountJSON(typing.TypedDict):
    mint: str
    owner: str
    amount: int
    delegate: typing.Optional[str]
    state: types.account_state.AccountStateJSON
    is_native: typing.Optional[int]
    delegated_amount: int
    close_authority: typing.Optional[str]


@dataclass
class Account:
    discriminator: typing.ClassVar = b"qB\xe06\xbcw\xf0e"
    layout: typing.ClassVar = borsh.CStruct(
        "mint" / BorshPubkey,
        "owner" / BorshPubkey,
        "amount" / borsh.U64,
        "delegate" / COption(BorshPubkey),
        "state" / types.account_state.layout,
        "is_native" / COption(borsh.U64),
        "delegated_amount" / borsh.U64,
        "close_authority" / COption(BorshPubkey),
    )
    mint: Pubkey
    owner: Pubkey
    amount: int
    delegate: typing.Optional[Pubkey]
    state: types.account_state.AccountStateKind
    is_native: typing.Optional[int]
    delegated_amount: int
    close_authority: typing.Optional[Pubkey]

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["Account"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["Account"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["Account"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "Account":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = Account.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            mint=dec.mint,
            owner=dec.owner,
            amount=dec.amount,
            delegate=dec.delegate,
            state=types.account_state.from_decoded(dec.state),
            is_native=dec.is_native,
            delegated_amount=dec.delegated_amount,
            close_authority=dec.close_authority,
        )

    def to_json(self) -> AccountJSON:
        return {
            "mint": str(self.mint),
            "owner": str(self.owner),
            "amount": self.amount,
            "delegate": (None if self.delegate is None else str(self.delegate)),
            "state": self.state.to_json(),
            "is_native": self.is_native,
            "delegated_amount": self.delegated_amount,
            "close_authority": (
                None if self.close_authority is None else str(self.close_authority)
            ),
        }

    @classmethod
    def from_json(cls, obj: AccountJSON) -> "Account":
        return cls(
            mint=Pubkey.from_string(obj["mint"]),
            owner=Pubkey.from_string(obj["owner"]),
            amount=obj["amount"],
            delegate=(
                None if obj["delegate"] is None else Pubkey.from_string(obj["delegate"])
            ),
            state=types.account_state.from_json(obj["state"]),
            is_native=obj["is_native"],
            delegated_amount=obj["delegated_amount"],
            close_authority=(
                None
                if obj["close_authority"] is None
                else Pubkey.from_string(obj["close_authority"])
            ),
        )



from pathlib import Path

from anchorpy import Event, EventParser, Idl, Program
from solders.pubkey import Pubkey


def test_event_parser() -> None:
    path = Path("tests/idls/events.json")
    raw = path.read_text()
    idl = Idl.from_json(raw)
    program = Program(
        idl, Pubkey.from_string("2dhGsWUzy5YKUsjZdLHLmkNpUDAXkNa9MYWsPc4Ziqzy")
    )
    logs = [
        "Program 2dhGsWUzy5YKUsjZdLHLmkNpUDAXkNa9MYWsPc4Ziqzy invoke [1]",
        "Program log: Instruction: Initialize",
        "Program data: YLjF84sCWpQFAAAAAAAAAAUAAABoZWxsbw==",
        "Program 2dhGsWUzy5YKUsjZdLHLmkNpUDAXkNa9MYWsPc4Ziqzy consumed 1019 of 1400000 compute units",
        "Program 2dhGsWUzy5YKUsjZdLHLmkNpUDAXkNa9MYWsPc4Ziqzy success",
    ]
    parser = EventParser(program.program_id, program.coder)
    evts = []
    parser.parse_logs(logs, lambda evt: evts.append(evt))
    assert len(evts) == 1
    events_coder = program.coder.events
    event_cls = events_coder.layouts["MyEvent"].datacls  # type: ignore
    expected_data = event_cls(
        data=5,
        label="hello",
    )
    expected_event = Event(name="MyEvent", data=expected_data)
    assert evts[0] == expected_event



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class UiAmountToAmountArgs(typing.TypedDict):
    ui_amount: str


layout = borsh.CStruct(borsh.String)


class UiAmountToAmountAccounts(typing.TypedDict):
    mint: Pubkey


def ui_amount_to_amount(
    args: UiAmountToAmountArgs,
    accounts: UiAmountToAmountAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["mint"], is_signer=False, is_writable=False)
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xad\xf3@\x04g\x1f84"
    encoded_args = layout.build(
        {
            "ui_amount": args["ui_amount"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from typing import Iterator, Optional
from typing import Union as TypingUnion

from genpy import (
    Assign,
    FromImport,
    Generable,
    Suite,
)
from genpy import Class as BrokenClass
from genpy import Function as FunctionOriginal


class Class(BrokenClass):
    def generate(self) -> Iterator[str]:
        bases = self.bases
        if not bases:
            bases = []

        yield "class {}({}):".format(self.name, ", ".join(bases))
        if self.attributes:
            for f in self.attributes:
                yield from ("    " + f_line for f_line in f.generate())
        else:
            yield "    pass"


class TypedParam(Generable):
    def __init__(self, name: str, type_: Optional[str]) -> None:
        self.name = name
        self.type = type_

    def generate(self) -> Iterator[str]:
        if self.type is None:
            yield self.name
        else:
            yield f"{self.name}: {self.type}"


class Break(Generable):
    def generate(self):
        yield "break"


class Continue(Generable):
    def generate(self):
        yield "continue"


class Union(Generable):
    def __init__(self, members: list[str]) -> None:
        self.members = members

    def generate(self) -> Iterator[str]:
        joined = ",".join(self.members)
        yield f"typing.Union[{joined}]"


class Tuple(Generable):
    def __init__(self, members: list[str]) -> None:
        self.members = members

    def generate(self) -> Iterator[str]:
        joined = ",".join(self.members)
        yield f"({joined},)"


class List(Generable):
    def __init__(self, members: list[str]) -> None:
        self.members = members

    def generate(self) -> Iterator[str]:
        joined = ",".join(self.members)
        yield f"[{joined}]"


class TupleTypeAlias(Generable):
    def __init__(self, name: str, members: list[str]) -> None:
        self.name = name
        self.members = members

    def generate(self) -> Iterator[str]:
        yield str(Assign(self.name, f"tuple{List(self.members)}"))


class StrDictEntry(Generable):
    def __init__(self, key: str, val: TypingUnion[str, "StrDict"]) -> None:
        self.key = key
        self.val = val

    def generate(self) -> Iterator[str]:
        yield f'"{self.key}": {self.val},'


class NamedArg(Generable):
    def __init__(self, key: str, val: str) -> None:
        self.key = key
        self.val = val

    def generate(self) -> Iterator[str]:
        yield f"{self.key}={self.val},"


class Call(Generable):
    def __init__(self, func: str, args: list[NamedArg]) -> None:
        self.func = func
        self.args = args

    def generate(self) -> Iterator[str]:
        formatted_args = "".join(str(arg) for arg in self.args)
        yield f"{self.func}({formatted_args})"


class StrDict(Generable):
    def __init__(self, items: list[StrDictEntry]) -> None:
        self.items = items

    def generate(self) -> Iterator[str]:
        formatted_items = "".join(str(item) for item in self.items)
        yield "{" + formatted_items + "}"


class IntDictEntry(Generable):
    def __init__(self, key: int, val: str) -> None:
        self.key = key
        self.val = val

    def generate(self) -> Iterator[str]:
        yield f"{self.key}: {self.val},"


class IntDict(Generable):
    def __init__(self, items: list[IntDictEntry]) -> None:
        self.items = items

    def generate(self) -> Iterator[str]:
        formatted_items = "".join(str(item) for item in self.items)
        yield "{" + formatted_items + "}"


class Function(FunctionOriginal):
    def __init__(
        self,
        name: str,
        args: list[TypedParam],
        body: Generable,
        return_type: str,
        decorators: tuple[str, ...] = (),
        is_async: bool = False,
    ) -> None:
        super().__init__(name, args, body, decorators)
        self.return_type = return_type
        self.is_async = is_async

    def generate(self) -> Iterator[str]:
        yield from self.decorators
        arg_strings = []
        for arg in self.args:
            annotation = "" if arg.type is None else f": {arg.type}"
            arg_strings.append(f"{arg.name}{annotation}")
        def_base = "async def" if self.is_async else "def"
        yield "{} {}({}) -> {}:".format(
            def_base, self.name, ", ".join(arg_strings), self.return_type
        )
        yield from self.body.generate()


class StaticMethod(Function):
    def __init__(
        self, name: str, args: list[TypedParam], body: Generable, return_type: str
    ) -> None:
        super().__init__(name, args, body, return_type, ("@staticmethod",))


class ClassMethod(Function):
    def __init__(
        self,
        name: str,
        extra_args: list[TypedParam],
        body: Generable,
        return_type: str,
        is_async: bool = False,
    ) -> None:
        args = [TypedParam("cls", None), *extra_args]
        super().__init__(name, args, body, return_type, ("@classmethod",), is_async)


class Method(Function):
    def __init__(
        self, name: str, extra_args: list[TypedParam], body: Generable, return_type: str
    ) -> None:
        args = [TypedParam("self", None), *extra_args]
        super().__init__(name, args, body, return_type)


class InitMethod(Method):
    def __init__(self, extra_args: list[TypedParam], body: Generable) -> None:
        super().__init__("__init__", extra_args, body, "None")


class Dataclass(Class):
    def __init__(
        self,
        name,
        attributes: list[TypingUnion[TypedParam, Assign, ClassMethod, Method]],
    ) -> None:
        super().__init__(name, None, attributes)

    def generate(self) -> Iterator[str]:
        yield "@dataclass"
        yield from super().generate()


class TypedDict(Class):
    def __init__(self, name, params: list[TypedParam]) -> None:
        super().__init__(name, ["typing.TypedDict"], params)

    def generate(self) -> Iterator[str]:
        yield from super().generate()


class Try(Generable):
    def __init__(self, try_body, to_catch, except_body):
        if not isinstance(try_body, Suite):
            try_body = Suite(try_body)
        if not isinstance(except_body, Suite):
            except_body = Suite(except_body)
        self.try_body = try_body
        self.to_catch = to_catch
        self.except_body = except_body

    def generate(self):
        yield "try:"
        yield from self.try_body.generate()
        yield f"except {self.to_catch}:"
        yield from self.except_body.generate()


ANNOTATIONS_IMPORT = FromImport("__future__", ["annotations"])



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from solders.sysvar import RENT, CLOCK
from solders.instruction import Instruction, AccountMeta
from ..program_id import PROGRAM_ID


class InitializeAccounts(typing.TypedDict):
    state: Pubkey
    payer: Pubkey


def initialize(
    accounts: InitializeAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=CLOCK, is_signer=False, is_writable=False),
        AccountMeta(pubkey=RENT, is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["payer"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xaf\xafm\x1f\r\x98\x9b\xed"
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



"""This module contains code for handling Anchor events."""
import binascii
from base64 import b64decode
from dataclasses import dataclass
from typing import Callable, List, Optional, cast

from solders.pubkey import Pubkey

from anchorpy.coder.coder import Coder
from anchorpy.program.common import Event

PROGRAM_LOG = "Program log: "
PROGRAM_DATA = "Program data: "
PROGRAM_LOG_START_INDEX = len(PROGRAM_LOG)
PROGRAM_DATA_START_INDEX = len(PROGRAM_DATA)


class _ExecutionContext:
    """Stack frame execution context, allowing one to track what program is executing for a given log."""  # noqa: E501

    def __init__(self, log: str) -> None:
        """Init.

        Args:
            log: The log to process.

        Raises:
            ValueError: If the log line is malformed.
        """
        try:
            program = log.split("Program ")[1].split(" invoke [")[0]
        except IndexError as e:
            raise ValueError("Could not find program invocation log line") from e
        self.stack = [program]

    def program(self) -> str:
        """Return the currently executing program.

        Returns:
            The name of the program.
        """
        return self.stack[-1]

    def push(self, new_program: str) -> None:
        """Add to the stack.

        Args:
            new_program: The program to add.
        """
        self.stack.append(new_program)

    def pop(self) -> None:
        """Pop from the stack."""
        self.stack.pop()


@dataclass
class EventParser:
    """Parser to handle on_logs callbacks."""

    program_id: Pubkey
    coder: Coder

    def parse_logs(self, logs: List[str], callback: Callable[[Event], None]) -> None:
        """Parse a list of logs using a provided callback.

        Args:
            logs: The logs to parse.
            callback: The function to handle the parsed log.
        """
        log_scanner = _LogScanner(logs)
        execution = _ExecutionContext(cast(str, log_scanner.to_next()))
        log = log_scanner.to_next()
        while log is not None:
            event, new_program, did_pop = self.handle_log(execution, log)
            if event is not None:
                callback(event)
            if new_program is not None:
                execution.push(new_program)
            if did_pop:
                execution.pop()
            log = log_scanner.to_next()

    def handle_log(
        self,
        execution: _ExecutionContext,
        log: str,
    ) -> tuple[Optional[Event], Optional[str], bool]:
        """Main log handler.

        Args:
            execution: The execution stack.
            log: log string from the RPC node.

        Returns:
            A three element array of the event, the next program
            that was invoked for CPI, and a boolean indicating if
            a program has completed execution (and thus should be popped off the
            execution stack).
        """
        # Executing program is this program.
        if execution.stack and execution.program() == str(self.program_id):
            return self.handle_program_log(log)
        # Executing program is not this program.
        return (None, *self.handle_system_log(log))

    def handle_program_log(
        self, log: str
    ) -> tuple[Optional[Event], Optional[str], bool]:
        """Handle logs from *this* program.

        Args:
            log: log string from the RPC node.

        """
        # This is a `msg!` log or a `sol_log_data!` log.
        if log.startswith(PROGRAM_LOG) or log.startswith(PROGRAM_DATA):
            log_str = (
                log[PROGRAM_LOG_START_INDEX:]
                if log.startswith(PROGRAM_LOG)
                else log[PROGRAM_DATA_START_INDEX:]
            )
            try:
                decoded = b64decode(log_str)
            except binascii.Error:
                return None, None, False
            event = self.coder.events.parse(decoded)
            return event, None, False
        return (None, *self.handle_system_log(log))

    def handle_system_log(self, log: str) -> tuple[Optional[str], bool]:
        """Handle logs when the current program being executing is *not* this.

        Args:
            log: log string from the RPC node.

        """
        log_start = log.split(":")[0]
        splitted = log_start.split(" ")
        invoke_msg = f"Program {str(self.program_id)} invoke"
        if len(splitted) == 3 and splitted[0] == "Program" and splitted[2] == "success":
            return None, True
        if log_start.startswith(invoke_msg):
            return str(self.program_id), False
        if "invoke" in log_start:
            return "cpi", False
        return None, False


@dataclass
class _LogScanner:
    """Object that iterates over logs."""

    logs: list[str]

    def to_next(self) -> Optional[str]:
        """Move to the next log item.

        Returns:
            The next log line, or None if there's nothing to return.
        """
        if self.logs:
            log = self.logs[0]
            self.logs = self.logs[1:]
            return log
        return None



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
from ..program_id import PROGRAM_ID


def cause_error(
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = []
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"Ch%\x11\x02\x9bD\x11"
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



import json

import pytest
from anchorpy import Coder
from anchorpy3_core.idl import Idl


@pytest.mark.unit
def test_can_encode_and_decode_user_defined_types():
    """Test that the TypesCoder can encode and decode user-defined types."""
    idl_json = {
        "version": "0.0.0",
        "name": "basic_0",
        "address": "Test111111111111111111111111111111111111111",
        "instructions": [
            {
                "name": "initialize",
                "accounts": [],
                "args": [],
                "discriminator": [],
            },
        ],
        "types": [
            {
                "name": "MintInfo",
                "type": {
                    "kind": "struct",
                    "fields": [
                        {
                            "name": "minted",
                            "type": "bool",
                        },
                        {
                            "name": "metadataUrl",
                            "type": "string",
                        },
                    ],
                },
            },
        ],
    }
    idl = Idl.from_json(json.dumps(idl_json))
    coder = Coder(idl)

    mint_info = {
        "minted": True,
        "metadata_url": "hello",
    }
    encoded = coder.types.encode("MintInfo", mint_info)
    decoded = coder.types.decode("MintInfo", encoded)

    # Compare decoded values with original
    assert decoded.minted == mint_info["minted"]
    assert decoded.metadata_url == mint_info["metadata_url"]


@pytest.mark.unit
def test_can_encode_and_decode_large_integers():
    """Test that the TypesCoder can encode and decode 128-bit integers."""
    idl_json = {
        "version": "0.0.0",
        "name": "basic_0",
        "address": "Test111111111111111111111111111111111111111",
        "instructions": [
            {
                "name": "initialize",
                "accounts": [],
                "args": [],
                "discriminator": [],
            },
        ],
        "types": [
            {
                "name": "IntegerTest",
                "type": {
                    "kind": "struct",
                    "fields": [
                        {
                            "name": "unsigned",
                            "type": "u128",
                        },
                        {
                            "name": "signed",
                            "type": "i128",
                        },
                    ],
                },
            },
        ],
    }
    idl = Idl.from_json(json.dumps(idl_json))
    coder = Coder(idl)

    integer_test = {
        "unsigned": 2588012355,
        "signed": -93842345,
    }
    encoded = coder.types.encode("IntegerTest", integer_test)
    decoded = coder.types.decode("IntegerTest", encoded)

    assert decoded.unsigned == integer_test["unsigned"]
    assert decoded.signed == integer_test["signed"]



"""This module provides `AccountsCoder` and `_account_discriminator`."""
from hashlib import sha256
from typing import Any, Tuple

from anchorpy3_core.idl import Idl
from construct import Adapter, Bytes, Container, Sequence, Switch

from anchorpy.coder.idl import _typedef_layout
from anchorpy.coder.idl_compat import get_account_discriminator, get_account_type_definition
from anchorpy.program.common import NamedInstruction as AccountToSerialize

ACCOUNT_DISCRIMINATOR_SIZE = 8  # bytes


class AccountsCoder(Adapter):
    """Encodes and decodes account data."""

    def __init__(self, idl: Idl) -> None:
        """Init.

        Args:
            idl: The parsed IDL object.
        """
        # Build account layouts - handling both old and new formats
        self._accounts_layout = {}
        for acc in idl.accounts:
            # Get account name (handle both object and dict)
            acc_name = acc.name if hasattr(acc, 'name') else acc.get('name')
            # Get the type definition (handles both old and new format)
            type_def = get_account_type_definition(acc, idl.types)
            self._accounts_layout[acc_name] = _typedef_layout(type_def, idl.types, acc_name)

        # Support both calculated and precomputed discriminators
        self.acc_name_to_discriminator = {}
        for acc in idl.accounts:
            # Get account name (handle both object and dict)
            acc_name = acc.name if hasattr(acc, 'name') else acc.get('name')
            # New format: use precomputed discriminator if available
            disc_list = get_account_discriminator(acc)
            if disc_list:
                disc = bytes(disc_list)
            else:
                # Old format: calculate from name
                disc = _account_discriminator(acc_name)
            self.acc_name_to_discriminator[acc_name] = disc
        self.discriminator_to_acc_name = {
            disc: acc_name for acc_name, disc in self.acc_name_to_discriminator.items()
        }
        discriminator_to_typedef_layout = {
            disc: self._accounts_layout[acc_name]
            for acc_name, disc in self.acc_name_to_discriminator.items()
        }
        subcon = Sequence(
            "discriminator" / Bytes(ACCOUNT_DISCRIMINATOR_SIZE),
            Switch(lambda this: this.discriminator, discriminator_to_typedef_layout),
        )
        super().__init__(subcon)  # type: ignore

    def decode(self, obj: bytes) -> Container[Any]:
        """Decode account data.

        Args:
            obj: Data to decode.

        Returns:
            Decoded data.
        """
        return self.parse(obj).data

    def _decode(self, obj: Tuple[bytes, Any], context, path) -> AccountToSerialize:
        return AccountToSerialize(
            data=obj[1],
            name=self.discriminator_to_acc_name[obj[0]],
        )

    def _encode(self, obj: AccountToSerialize, context, path) -> Tuple[bytes, Any]:
        discriminator = self.acc_name_to_discriminator[obj.name]
        return discriminator, obj.data


def _account_discriminator(name: str) -> bytes:
    """Calculate unique 8 byte discriminator prepended to all anchor accounts.

    Args:
        name: The account name.

    Returns:
        The discriminator in bytes.
    """
    return sha256(f"account:{name}".encode()).digest()[:ACCOUNT_DISCRIMINATOR_SIZE]



"""Common utilities."""
from dataclasses import dataclass
from typing import Any, Dict, NamedTuple, Tuple, Union, cast

from anchorpy3_core.idl import (
    IdlAccountItem,
    IdlAccounts,
    IdlInstruction,
)
from construct import Container
from pyheck import snake
from solders.pubkey import Pubkey

from anchorpy.program.context import Accounts

AddressType = Union[Pubkey, str]


class Event(NamedTuple):
    """A parsed event object."""

    name: str
    data: Any


@dataclass
class NamedInstruction:
    """Container for a named instruction.

    Attributes:
        data: The actual instruction data.
        name: The name of the instruction.
    """

    data: Union[Dict[str, Any], Container[Any]]
    name: str


def _to_instruction(idl_ix: IdlInstruction, args: Tuple) -> NamedInstruction:
    """Convert an IDL instruction and arguments to an Instruction object.

    Args:
        idl_ix: The IDL instruction object.
        args: The instruction arguments.

    Raises:
        ValueError: If the incorrect number of arguments is provided.

    Returns:
        The parsed Instruction object.
    """
    if len(idl_ix.args) != len(args):
        raise ValueError("Invalid argument length")
    ix: Dict[str, Any] = {}
    for idx, ix_arg in enumerate(idl_ix.args):
        ix[snake(ix_arg.name)] = args[idx]
    return NamedInstruction(data=ix, name=snake(idl_ix.name))


def validate_accounts(ix_accounts: list[IdlAccountItem], accounts: Accounts):
    """Check that accounts passed in `ctx` match the IDL.

    Args:
        ix_accounts: Accounts from the IDL.
        accounts: Accounts from the `ctx` arg.

    Raises:
        ValueError: If `ctx` accounts don't match the IDL.
    """
    for acc in ix_accounts:
        acc_name = snake(acc.name)
        if isinstance(acc, IdlAccounts):
            nested = cast(Accounts, accounts[acc_name])
            validate_accounts(acc.accounts, nested)
        elif acc_name not in accounts:
            raise ValueError(f"Invalid arguments: {acc_name} not provided")


def translate_address(address: AddressType) -> Pubkey:
    """Convert `str | Pubkey` into `Pubkey`.

    Args:
        address: Public key as string or `Pubkey`.

    Returns:
        Public key as `Pubkey`.
    """
    if isinstance(address, str):
        return Pubkey.from_string(address)
    return address



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from solders.instruction import Instruction, AccountMeta
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class SetupGameArgs(typing.TypedDict):
    player_two: Pubkey


layout = borsh.CStruct("player_two" / BorshPubkey)


class SetupGameAccounts(typing.TypedDict):
    game: Pubkey
    player_one: Pubkey


def setup_game(
    args: SetupGameArgs,
    accounts: SetupGameAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["game"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=accounts["player_one"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xb4\xda\x80K:\xde#R"
    encoded_args = layout.build(
        {
            "player_two": args["player_two"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.sysvar import RENT
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class InitializeMultisigArgs(typing.TypedDict):
    m: int


layout = borsh.CStruct("m" / borsh.U8)


class InitializeMultisigAccounts(typing.TypedDict):
    multisig: Pubkey


def initialize_multisig(
    args: InitializeMultisigArgs,
    accounts: InitializeMultisigAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["multisig"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=RENT, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xdc\x82u\x15\x1b\xe3N\xd5"
    encoded_args = layout.build(
        {
            "m": args["m"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class ApproveCheckedArgs(typing.TypedDict):
    amount: int
    decimals: int


layout = borsh.CStruct("amount" / borsh.U64, "decimals" / borsh.U8)


class ApproveCheckedAccounts(typing.TypedDict):
    source: Pubkey
    mint: Pubkey
    delegate: Pubkey
    owner: Pubkey


def approve_checked(
    args: ApproveCheckedArgs,
    accounts: ApproveCheckedAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["source"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["delegate"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["owner"], is_signer=True, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"/\xc5\xfe*:\xc9:m"
    encoded_args = layout.build(
        {
            "amount": args["amount"],
            "decimals": args["decimals"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



import typing
from dataclasses import dataclass
from solders.pubkey import Pubkey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey, COption
from ..program_id import PROGRAM_ID


class MintJSON(typing.TypedDict):
    mint_authority: typing.Optional[str]
    supply: int
    decimals: int
    is_initialized: bool
    freeze_authority: typing.Optional[str]


@dataclass
class Mint:
    discriminator: typing.ClassVar = b"P\xbc\xf5\x14_\x8a9\x9c"
    layout: typing.ClassVar = borsh.CStruct(
        "mint_authority" / COption(BorshPubkey),
        "supply" / borsh.U64,
        "decimals" / borsh.U8,
        "is_initialized" / borsh.Bool,
        "freeze_authority" / COption(BorshPubkey),
    )
    mint_authority: typing.Optional[Pubkey]
    supply: int
    decimals: int
    is_initialized: bool
    freeze_authority: typing.Optional[Pubkey]

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["Mint"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["Mint"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["Mint"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "Mint":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = Mint.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            mint_authority=dec.mint_authority,
            supply=dec.supply,
            decimals=dec.decimals,
            is_initialized=dec.is_initialized,
            freeze_authority=dec.freeze_authority,
        )

    def to_json(self) -> MintJSON:
        return {
            "mint_authority": (
                None if self.mint_authority is None else str(self.mint_authority)
            ),
            "supply": self.supply,
            "decimals": self.decimals,
            "is_initialized": self.is_initialized,
            "freeze_authority": (
                None if self.freeze_authority is None else str(self.freeze_authority)
            ),
        }

    @classmethod
    def from_json(cls, obj: MintJSON) -> "Mint":
        return cls(
            mint_authority=(
                None
                if obj["mint_authority"] is None
                else Pubkey.from_string(obj["mint_authority"])
            ),
            supply=obj["supply"],
            decimals=obj["decimals"],
            is_initialized=obj["is_initialized"],
            freeze_authority=(
                None
                if obj["freeze_authority"] is None
                else Pubkey.from_string(obj["freeze_authority"])
            ),
        )



#!/usr/bin/env python3
"""Test parsing loopscale_v2.json step by step to find the issue."""

import json
from pathlib import Path
from anchorpy3_core.idl import Idl

idl_path = Path(__file__).parent.parent / "loopscale_v2.json"

# First, let's load it as pure JSON
print("Loading as JSON...")
with open(idl_path) as f:
    idl_json = json.load(f)
print("✅ JSON loads successfully")

# Now let's try to parse it with anchorpy_core
print("\nParsing with anchorpy_core.idl.Idl...")

# Try parsing with more detailed error handling
json_str = json.dumps(idl_json)
print(f"JSON string length: {len(json_str)}")

try:
    idl_obj = Idl.from_json(json_str)
    print("✅ Successfully parsed!")
except Exception as e:
    print(f"❌ Failed: {e}")

    # Try to narrow down the issue
    print("\nTrying to narrow down the issue...")

    # Test with minimal IDL
    minimal_idl = {
        "version": idl_json.get("version"),
        "name": idl_json.get("name"),
        "address": idl_json.get("address"),
        "instructions": [],
        "accounts": [],
        "types": [],
        "errors": []
    }

    try:
        print("Testing minimal IDL...")
        Idl.from_json(json.dumps(minimal_idl))
        print("✅ Minimal IDL works")
    except Exception as e2:
        print(f"❌ Even minimal IDL fails: {e2}")

    # Add components one by one
    if idl_json.get("metadata"):
        minimal_idl["metadata"] = idl_json["metadata"]
        try:
            print("Testing with metadata...")
            Idl.from_json(json.dumps(minimal_idl))
            print("✅ Metadata works")
        except Exception as e2:
            print(f"❌ Metadata causes issue: {e2}")

    # Add accounts
    minimal_idl["accounts"] = idl_json.get("accounts", [])
    try:
        print("Testing with accounts...")
        Idl.from_json(json.dumps(minimal_idl))
        print("✅ Accounts work")
    except Exception as e2:
        print(f"❌ Accounts cause issue: {e2}")

    # Add types one by one
    print("\nTesting types one by one...")
    minimal_idl["accounts"] = []  # Remove accounts for now
    for i, type_def in enumerate(idl_json.get("types", [])):
        minimal_idl["types"] = idl_json["types"][:i+1]
        try:
            Idl.from_json(json.dumps(minimal_idl))
        except Exception as e2:
            print(f"❌ Type {i} ({type_def.get('name')}) causes issue: {e2}")
            print(f"   Type definition: {json.dumps(type_def, indent=2)}")
            break
    else:
        print(f"✅ All {len(idl_json.get('types', []))} types work individually")

    # Add instructions
    minimal_idl["types"] = []  # Clear types
    minimal_idl["instructions"] = idl_json.get("instructions", [])
    try:
        print("\nTesting with instructions...")
        Idl.from_json(json.dumps(minimal_idl))
        print("✅ Instructions work")
    except Exception as e2:
        print(f"❌ Instructions cause issue: {e2}")

    # Add errors
    minimal_idl["instructions"] = []  # Clear instructions
    minimal_idl["errors"] = idl_json.get("errors", [])
    try:
        print("Testing with errors...")
        Idl.from_json(json.dumps(minimal_idl))
        print("✅ Errors work")
    except Exception as e2:
        print(f"❌ Errors cause issue: {e2}")

    # Now test everything together
    print("\nTesting full IDL again...")
    full_test = {
        "version": idl_json.get("version"),
        "name": idl_json.get("name"),
        "address": idl_json.get("address"),
        "metadata": idl_json.get("metadata"),
        "instructions": idl_json.get("instructions", []),
        "accounts": idl_json.get("accounts", []),
        "types": idl_json.get("types", []),
        "errors": idl_json.get("errors", [])
    }

    try:
        Idl.from_json(json.dumps(full_test))
        print("✅ Full IDL works when reconstructed")
    except Exception as e2:
        print(f"❌ Full IDL still fails: {e2}")


from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class InitMyAccountArgs(typing.TypedDict):
    seed_a: int


layout = borsh.CStruct("seed_a" / borsh.U8)
NESTED_NESTED_ACCOUNT_NESTED = Pubkey.find_program_address(
    seeds=[
        b"nested-seed",
        b"test",
        b"hi",
        b"hi",
        b"\x01",
        b"\x02\x00\x00\x00",
        b"\x03\x00\x00\x00\x00\x00\x00\x00",
    ],
    program_id=PROGRAM_ID,
)[0]
INIT_MY_ACCOUNT_ACCOUNTS_ACCOUNT = Pubkey.find_program_address(
    seeds=[
        b"another-seed",
        b"test",
        b"hi",
        b"hi",
        b"\x01",
        b"\x02\x00\x00\x00",
        b"\x03\x00\x00\x00\x00\x00\x00\x00",
    ],
    program_id=PROGRAM_ID,
)[0]


class InitMyAccountAccounts(typing.TypedDict):
    base: Pubkey
    base2: Pubkey


def init_my_account(
    args: InitMyAccountArgs,
    accounts: InitMyAccountAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["base"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["base2"], is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=INIT_MY_ACCOUNT_ACCOUNTS_ACCOUNT, is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=NESTED_NESTED_ACCOUNT_NESTED, is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b" ;\x05\xcd^E\xe3z"
    encoded_args = layout.build(
        {
            "seed_a": args["seed_a"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from pathlib import Path

from anchorpy import Idl, InstructionCoder
from anchorpy.program.common import _to_instruction
from anchorpy.program.context import _check_args_length
from pytest import mark


@mark.unit
def test_instruction_coder() -> None:
    """Test InstructionCoder behaves as expected."""
    raw = Path("tests/idls/basic_1.json").read_text()
    idl = Idl.from_json(raw)
    idl_ix = idl.instructions[0]
    args = (1234,)
    _check_args_length(idl_ix, args)
    ix = _to_instruction(idl_ix, args)
    coder = InstructionCoder(idl)
    encoded = coder.build(ix)
    assert encoded == b"\xaf\xafm\x1f\r\x98\x9b\xed\xd2\x04\x00\x00\x00\x00\x00\x00"
    assert coder.parse(encoded) == ix



from pathlib import Path

from black import FileMode, format_str
from genpy import Assign, Collection, FromImport


def gen_program_id_code(program_id: str) -> str:
    import_line = FromImport("solders.pubkey", ["Pubkey"])
    assignment_line = Assign("PROGRAM_ID", f'Pubkey.from_string("{program_id}")')
    return str(Collection([import_line, assignment_line]))


def gen_program_id(program_id: str, root: Path) -> None:
    code = gen_program_id_code(program_id)
    formatted = format_str(code, mode=FileMode())
    (root / "program_id.py").write_text(formatted)



import typing
from dataclasses import dataclass
from solders.pubkey import Pubkey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID


class MultisigJSON(typing.TypedDict):
    m: int
    n: int
    is_initialized: bool
    signers: list[str]


@dataclass
class Multisig:
    discriminator: typing.ClassVar = b"\xe0ty\xbaD\xa1O\xec"
    layout: typing.ClassVar = borsh.CStruct(
        "m" / borsh.U8,
        "n" / borsh.U8,
        "is_initialized" / borsh.Bool,
        "signers" / BorshPubkey[11],
    )
    m: int
    n: int
    is_initialized: bool
    signers: list[Pubkey]

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["Multisig"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["Multisig"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["Multisig"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "Multisig":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = Multisig.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            m=dec.m,
            n=dec.n,
            is_initialized=dec.is_initialized,
            signers=dec.signers,
        )

    def to_json(self) -> MultisigJSON:
        return {
            "m": self.m,
            "n": self.n,
            "is_initialized": self.is_initialized,
            "signers": list(map(lambda item: str(item), self.signers)),
        }

    @classmethod
    def from_json(cls, obj: MultisigJSON) -> "Multisig":
        return cls(
            m=obj["m"],
            n=obj["n"],
            is_initialized=obj["is_initialized"],
            signers=list(map(lambda item: Pubkey.from_string(item), obj["signers"])),
        )



import typing
from anchorpy.error import ProgramError


class NotRentExempt(ProgramError):
    def __init__(self) -> None:
        super().__init__(0, "Lamport balance below rent-exempt threshold")

    code = 0
    name = "NotRentExempt"
    msg = "Lamport balance below rent-exempt threshold"


class InsufficientFunds(ProgramError):
    def __init__(self) -> None:
        super().__init__(1, "Insufficient funds")

    code = 1
    name = "InsufficientFunds"
    msg = "Insufficient funds"


class InvalidMint(ProgramError):
    def __init__(self) -> None:
        super().__init__(2, "Invalid Mint")

    code = 2
    name = "InvalidMint"
    msg = "Invalid Mint"


class MintMismatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(3, "Account not associated with this Mint")

    code = 3
    name = "MintMismatch"
    msg = "Account not associated with this Mint"


class OwnerMismatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(4, "Owner does not match")

    code = 4
    name = "OwnerMismatch"
    msg = "Owner does not match"


class FixedSupply(ProgramError):
    def __init__(self) -> None:
        super().__init__(5, "Fixed supply")

    code = 5
    name = "FixedSupply"
    msg = "Fixed supply"


class AlreadyInUse(ProgramError):
    def __init__(self) -> None:
        super().__init__(6, "Already in use")

    code = 6
    name = "AlreadyInUse"
    msg = "Already in use"


class InvalidNumberOfProvidedSigners(ProgramError):
    def __init__(self) -> None:
        super().__init__(7, "Invalid number of provided signers")

    code = 7
    name = "InvalidNumberOfProvidedSigners"
    msg = "Invalid number of provided signers"


class InvalidNumberOfRequiredSigners(ProgramError):
    def __init__(self) -> None:
        super().__init__(8, "Invalid number of required signers")

    code = 8
    name = "InvalidNumberOfRequiredSigners"
    msg = "Invalid number of required signers"


class UninitializedState(ProgramError):
    def __init__(self) -> None:
        super().__init__(9, "State is unititialized")

    code = 9
    name = "UninitializedState"
    msg = "State is unititialized"


class NativeNotSupported(ProgramError):
    def __init__(self) -> None:
        super().__init__(10, "Instruction does not support native tokens")

    code = 10
    name = "NativeNotSupported"
    msg = "Instruction does not support native tokens"


class NonNativeHasBalance(ProgramError):
    def __init__(self) -> None:
        super().__init__(
            11, "Non-native account can only be closed if its balance is zero"
        )

    code = 11
    name = "NonNativeHasBalance"
    msg = "Non-native account can only be closed if its balance is zero"


class InvalidInstruction(ProgramError):
    def __init__(self) -> None:
        super().__init__(12, "Invalid instruction")

    code = 12
    name = "InvalidInstruction"
    msg = "Invalid instruction"


class InvalidState(ProgramError):
    def __init__(self) -> None:
        super().__init__(13, "State is invalid for requested operation")

    code = 13
    name = "InvalidState"
    msg = "State is invalid for requested operation"


class Overflow(ProgramError):
    def __init__(self) -> None:
        super().__init__(14, "Operation overflowed")

    code = 14
    name = "Overflow"
    msg = "Operation overflowed"


class AuthorityTypeNotSupported(ProgramError):
    def __init__(self) -> None:
        super().__init__(15, "Account does not support specified authority type")

    code = 15
    name = "AuthorityTypeNotSupported"
    msg = "Account does not support specified authority type"


class MintCannotFreeze(ProgramError):
    def __init__(self) -> None:
        super().__init__(16, "This token mint cannot freeze accounts")

    code = 16
    name = "MintCannotFreeze"
    msg = "This token mint cannot freeze accounts"


class AccountFrozen(ProgramError):
    def __init__(self) -> None:
        super().__init__(17, "Account is frozen")

    code = 17
    name = "AccountFrozen"
    msg = "Account is frozen"


class MintDecimalsMismatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(
            18, "The provided decimals value different from the Mint decimals"
        )

    code = 18
    name = "MintDecimalsMismatch"
    msg = "The provided decimals value different from the Mint decimals"


class NonNativeNotSupported(ProgramError):
    def __init__(self) -> None:
        super().__init__(19, "Instruction does not support non-native tokens")

    code = 19
    name = "NonNativeNotSupported"
    msg = "Instruction does not support non-native tokens"


CustomError = typing.Union[
    NotRentExempt,
    InsufficientFunds,
    InvalidMint,
    MintMismatch,
    OwnerMismatch,
    FixedSupply,
    AlreadyInUse,
    InvalidNumberOfProvidedSigners,
    InvalidNumberOfRequiredSigners,
    UninitializedState,
    NativeNotSupported,
    NonNativeHasBalance,
    InvalidInstruction,
    InvalidState,
    Overflow,
    AuthorityTypeNotSupported,
    MintCannotFreeze,
    AccountFrozen,
    MintDecimalsMismatch,
    NonNativeNotSupported,
]
CUSTOM_ERROR_MAP: dict[int, CustomError] = {
    0: NotRentExempt(),
    1: InsufficientFunds(),
    2: InvalidMint(),
    3: MintMismatch(),
    4: OwnerMismatch(),
    5: FixedSupply(),
    6: AlreadyInUse(),
    7: InvalidNumberOfProvidedSigners(),
    8: InvalidNumberOfRequiredSigners(),
    9: UninitializedState(),
    10: NativeNotSupported(),
    11: NonNativeHasBalance(),
    12: InvalidInstruction(),
    13: InvalidState(),
    14: Overflow(),
    15: AuthorityTypeNotSupported(),
    16: MintCannotFreeze(),
    17: AccountFrozen(),
    18: MintDecimalsMismatch(),
    19: NonNativeNotSupported(),
}


def from_code(code: int) -> typing.Optional[CustomError]:
    maybe_err = CUSTOM_ERROR_MAP.get(code)
    if maybe_err is None:
        return None
    return maybe_err



"""This module contains code for creating simulate functions."""
from typing import Any, Awaitable, Dict, NamedTuple, Protocol

from anchorpy3_core.idl import Idl, IdlInstruction
from solana.rpc.commitment import Confirmed
from solana.rpc.core import RPCException
from solders.pubkey import Pubkey

from anchorpy.coder.coder import Coder
from anchorpy.error import ProgramError
from anchorpy.program.context import EMPTY_CONTEXT, Context, _check_args_length
from anchorpy.program.event import Event, EventParser
from anchorpy.program.namespace.transaction import _TransactionFn
from anchorpy.provider import Provider


class SimulateResponse(NamedTuple):
    """The result of a simulate function call."""

    events: list[Event]
    raw: list[str]


class _SimulateFn(Protocol):
    """A single method generated from an IDL.

    It simulates a method against a cluster configured by the provider,
    returning a list of all the events and raw logs that were emitted
    during the execution of the method.
    """

    def __call__(
        self,
        *args: Any,
        ctx: Context = EMPTY_CONTEXT,
    ) -> Awaitable[SimulateResponse]:
        """Protocol definition.

        Args:
            *args: The positional arguments for the program. The type and number
                of these arguments depend on the program being used.
            ctx: non-argument parameters to pass to the method.

        """


def _build_simulate_item(
    idl_ix: IdlInstruction,
    tx_fn: _TransactionFn,
    idl_errors: Dict[int, str],
    provider: Provider,
    coder: Coder,
    program_id: Pubkey,
    idl: Idl,
) -> _SimulateFn:
    """Build the function to simulate transactions for a given method of a program.

    Args:
        idl_ix: An IDL instruction object.
        tx_fn: The function to generate the `Transaction` object.
        idl_errors: Mapping of error code to message.
        provider: A provider instance.
        coder: The program's coder object.
        program_id: The program ID.
        idl: The parsed Idl instance.

    Returns:
        The simulate function.
    """

    async def simulate_fn(*args: Any, ctx: Context = EMPTY_CONTEXT) -> SimulateResponse:
        blockhash = (
            await provider.connection.get_latest_blockhash(Confirmed)
        ).value.blockhash
        tx = tx_fn(*args, payer=provider.wallet.payer, blockhash=blockhash, ctx=ctx)
        _check_args_length(idl_ix, args)
        resp = (await provider.simulate(tx, ctx.options)).value
        resp_err = resp.err
        logs = resp.logs or []
        if resp_err is None:
            events = []
            if idl.events is not None:
                parser = EventParser(program_id, coder)
                parser.parse_logs(logs, lambda evt: events.append(evt))
            return SimulateResponse(events, logs)
        else:
            translated_err = ProgramError.parse_tx_error(
                resp_err, idl_errors, program_id, logs
            )
            if translated_err is not None:
                raise translated_err
            raise RPCException(resp_err)

    return simulate_fn
